# *****************************************************************************
# * Copyright (c) 2013-2014, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial implementation
# *****************************************************************************

import atexit
import threading
import os.path
import sys
import time

import tcf as tcf  # @UnresolvedImport
import tcf.protocol as protocol  # @UnresolvedImport

import tcf.services.processes as processes  # @UnresolvedImport
import tcf.services.processes_v1 as processes_v1  # @UnresolvedImport
import tcf.services.runcontrol as runcontrol  # @UnresolvedImport
import tcf.services.streams as streams  # @UnresolvedImport


class TcfProtocolLogger (object):
    """A class to override TCF protocol's default logger.

    As we do not want TCF errors to be logged on the console, this logger
    simply does nothing on TCF protocol log messages.
    """

    def log(self, msg, x):
        """Logs the given message.

        :param msg: log entry text.
        :param x: an exception associated with the log entry or **None**.
        """
        pass

    def __del__(self):
        """Sometimes we get some protocol warnings. I would like to remove them
        by flushing the log cache here.
        """
        pass


class TcfValue(object):
    """TCF value container.

    A simple class to handle the value returned by an asynchronous TCF request.

    :param value: A value to initialise this TCF value class with.
    """
    def __init__(self, value=None):
        self._value = value

    def getValue(self):
        """Get this TCF stored value.

        If this TCF value has been set using the set() method, or at object
        creation time, the so-set value is returned. Else, **None** is
        returned.

        :returns: This TCF stored value, or **None**.
        """
        return (self._value)

    def setValue(self, value):
        """Set this TCF value.

        :param value: TCF value to set.

        :returns: **None**, always.
        """
        self._value = value


class StreamsListener(streams.StreamsListener):
    """A TCF streams service listener.

    :param service: streams service to listen to.
    """
    def __init__(self, service):
        self._service = service

    def created(self, streamType, streamID, contextID):
        """Called when a new stream is created.

        :param streamType: source type of the stream.
        :param streamID: ID of the stream.
        :param contextID: a context ID that is associated with the stream,
                          or **None**. Exact meaning of the context ID depends
                          on stream type. Stream types and context IDs are
                          defined by services that use Streams service to
                          transmit data.
        """
        class DoneRead(streams.DoneRead):
            """Call back interface for StreamsService.read() command.

            :param service: The streams service to read data from.
            :param streamID: The stream ID to read data from.
            :param size: Size of data to read on stream.
            """
            def __init__(self, service, streamID, size):
                self._service = service
                self._streamID = streamID
                self._size = size

            def doneRead(self, token, error, lost_size, data, eos):
                """Called when StreamsService.read() command is done.

                :param token: command handle.
                :param error: error object or **None**.
                :param lost_size: number of bytes that were lost because of
                                  buffer overflow. A *lost_size* of **-1**
                                  means unknown number of bytes were lost. If
                                  both *lost_size* and *data.length* are
                                  non-zero then lost bytes are considered
                                  located right before read bytes.
                :param data: bytes read from the stream.
                :param eos: true if end of stream was reached.
                """
                if data:
                    if isinstance(data, bytearray):
                        s = data.decode('utf-8')
                    else:
                        s = str(data)
                    sys.stdout.write(s)
                if not eos:
                    self._service.read(self._streamID, self._size, self)

        self._service.read(streamID, 4096, DoneRead(self._service, streamID,
                                                    4096))

    def disposed(self, streamType, streamID):
        """Called when a stream is disposed.

        :param streamType: source type of the stream.
        :param streamID: ID of the stream.
        """
        pass


def getService(connection, name):
    """Get a service proxy.

    The service proxy named *name* is retrieved from *connection*. If it
    exists (returned value is not **None**), it is then possible to send it
    TCF requests.

    :param connection: The connection to get service proxy from.
    :param name: The name of the service proxy to get from *connection*.

    :returns: A service proxy or **None**.
    """
    def callGetService(connection, condition, val):
        """Asynchronous request to get service proxy.

        :param connection: The connection to get service proxy from.
        :param condition: A threading.Condition the caller is pending on.
                          Caller is released from waiting through a
                          Condition.notify() call.
        :param val: A TcfValue handling the request returned value.

        :returns: **None**, always.
        """
        svc = connection.getRemoteService(name)
        val.setValue(svc)
        with condition:
            condition.notify()

    # create a condition to wait on, and a value to get the service proxy

    lock = threading.Condition()
    value = TcfValue()

    with lock:
        # Asynchronously call for the callGetService function.
        protocol.invokeLater(callGetService, connection=connection,
                             condition=lock, val=value)
        lock.wait(5)

    # Return TCF service proxy. May be None on timeout or missing service.

    return (value.getValue())


def getChildren(service, contextID=None):
    """Get a TCF context IDs from a given service.

    As many TCF services have a **getChildren()** command, this function is
    intended to implement a service-independant **getChildren()** command.

    :param service: The TCF service to get context list from.
    :param contextID: parent ID of the context list to get from *service*.

    :returns: A tuple of context IDs. Tuple may be empty on error, or if
              *contextID* does not have children.
    """
    def callGetChildren(service, condition, val):
        """Asynchronous request to get context children.

        :param service: The TCF service proxy to send request to.
        :param condition: A threading.Condition the caller is pending on.
                          Caller is released from waiting through a
                          Condition.notify() call.
        :param val: A TcfValue handling the request returned value.

        :returns: **None**, always.
        """
        class DoneGetChildren(object):
            """Client callback class for <service>.getChildren() command."""
            def doneGetChildren(self, token, error, ids):
                """Called when context list retrieval is done.

                :param token: pending command handle.
                :param error: error description if operation failed, **None**
                              if succeeded.
                :param context_ids: array of available context IDs.
                """
                if error:
                    protocol.log("Error from " + service.getName() +
                                 ".getContext()", error)
                else:
                    val.setValue(ids)
                with condition:
                    condition.notify()

        # start the process itself

        service.getChildren(contextID, DoneGetChildren())

    # create a condition to wait on, and a value to get the children ids

    lock = threading.Condition()
    value = TcfValue()

    with lock:
        # TCF requests must be called by the dispatch thread, wait for a
        # maximum of 10 seconds
        protocol.invokeLater(callGetChildren, service=service, condition=lock,
                             val=value)
        lock.wait(10)

    # Return the retrieved children IDs, or an empty tuple

    return (tuple(value.getValue() or []))


def getContext(service, contextID):
    """Get a TCF context from a given service.

    As most of the TCF services have a **getContext()** command, this
    function is intended to define a generic **getContext()** call which
    can address any service.

    As the function returns a context object, it is up to the caller to use it
    appropriately.

    For example, for the runcontrol service, check if the runcontrol context is
    a container :

    .. code-block:: python

        import tcf
        import tcf.services.runcontrol as runcontrol

        c = tcf.connect("TCP:127.0.0.1:1534")
        rcSvc = getService(c, runcontrol.NAME)
        rcIDs = getChildren(rcSvc, None)
        rcContext = getContext(rcSvc, rcIDs[0])

        print('Runcontrol context is a container: ' +
              str(rcContext.isContainer()))

    :param service: The TCF service to get context from.
    :param contextID: ID of the context to get from *service*

    :returns: A context properties dictionnary, or **None** on error.
    """
    def callGetContext(service, condition, val):
        """Asynchronous request to get context properties.

        :param service: The TCF service proxy to send request to.
        :param condition: A threading.Condition the caller is pending on.
                          Caller is released from waiting through a
                          Condition.notify() call.
        :param val: A TcfValue handling the request returned value.

        :returns: **None**, always.
        """
        class DoneGetContext(object):
            """Client callback class for <service>.getContext() command."""
            def doneGetContext(self, token, error, context):
                """Called when context data retrieval is done.

                :param token: pending command handle.
                :param error: error description if operation failed, **None**
                              if succeeded.
                :param context: context data.
                """
                if error:
                    protocol.log("Error from " + service.getName() +
                                 ".getContext()", error)
                else:
                    val.setValue(context)
                with condition:
                    condition.notify()

        # start the process itself

        service.getContext(contextID, DoneGetContext())

    # create a condition to wait on, and a value to get the children ids

    lock = threading.Condition()
    value = TcfValue()

    with lock:
        # TCF requests must be called by the dispatch thread, wait for a
        # maximum of 10 seconds
        protocol.invokeLater(callGetContext, service=service, condition=lock,
                             val=value)
        lock.wait(10)

    # Return the context properties, or None on error

    return (value.getValue())


def resume(context):
    """Resume a runcontrol context.

    The given *context* should be a RunControlContext, so that its resume()
    method may be called.

    :param context: A runcontrol context to resume.
    :type process: RunControlContext

    :returns: **None**, always.
    """
    def callResume(context, condition):
        """Asynchronous request to resume runcontrol context.

        :param context: The TCF RunControlContext to resume.
        :param condition: A threading.Condition the caller is pending on.
                          Caller is released from waiting through a
                          Condition.notify() call.

        :returns: **None**, always.
        """
        class DoneResume(runcontrol.DoneCommand):
            """Client call back interface for RunControlContext.resume()."""
            def doneCommand(self, token, error):
                """Called when run control command execution is complete.

                :param token: pending command handle.
                :param error: command execution error or **None**.
                """
                if error:
                    protocol.log("Error from RunContext.resume", error)
                with condition:
                    condition.notify()

        # Resume context with RM_RESUME mode, 1 time. No resume properties.

        context.resume(runcontrol.RM_RESUME, 1, {}, DoneResume())

    # create a condition to wait on

    lock = threading.Condition()

    with lock:
        # TCF requests must be called by the dispatch thread, wait for a
        # maximum of 10 seconds

        protocol.invokeLater(callResume, context=context, condition=lock)
        lock.wait(10)


def start(connection, path, *args):
    """Start a new process in suspended mode for the given connection.

    :param connection: The TCF connection to use services from.
    :param path: Path of the executable to start.
    :param args: command line arguments.
    """
    def callStart(connection, condition, path, arguments, val):
        """Asynchronous request to start a process.

        :param connection: The TCF connection to get processes service from.
        :param condition: A threading.Condition the caller is pending on.
                          Caller is released from waiting through a
                          Condition.notify() call.
        :param path: Path of the process to start on the target
        :param arguments: A list of program arguments for *path*
        :param val: A TcfValue handling the request returned value.

        :returns: **None**, always.
        """
        # get connection's processes service

        proc = connection.getRemoteService(processes_v1.NAME) or \
            connection.getRemoteService(processes.NAME)

        if not proc:
            with condition:
                print('No processes service available')
                condition.notify()
            return

        class DoneStart(processes.DoneStart):
            """Client callback interface for ProcessesService.start()."""

            def doneStart(self, token, error, process):
                """Called when process start is done.

                :param token: pending command handle.
                :param error: error description if operation failed, **None**
                              if  succeeded.
                :param process: ProcessContext object representing the started
                                process.
                """
                if error:
                    protocol.log("Error from Processes.start", error)
                else:
                    val.setValue(process)
                with condition:
                    condition.notify()

        # depending on the service, the start method only does 'doAttach', or
        # take a dictionnary of options

        if (proc.getName() == processes_v1.NAME):
            opts = {processes_v1.START_ATTACH: True,
                    processes_v1.START_USE_TERMINAL: True}
        else:
            opts = True

        # start the process itself

        cmdLine = [path]
        if arguments:
            cmdLine += arguments

        proc.start(os.path.dirname(path), path, cmdLine, None, opts,
                   DoneStart())

    # create a condition to wait on, and a value to get the children ids

    lock = threading.Condition()
    value = TcfValue()

    with lock:
        # TCF requests must be called by the dispatch thread, wait for a
        # maximum of 10 seconds
        protocol.invokeLater(callStart, connection=connection, condition=lock,
                             path=path, arguments=args, val=value)
        lock.wait(10)

    return (value.getValue())


def state(context):
    """Get the state of a RunControlContext.

    Getting the state of a RunControlContext is asynchronous.

    :param context: The context to get state for.
    :type context: RunControlContext

    :returns: A tuple of four elements representing the RunControl context
              state :

                  - A boolean stating if the context id suspended or not
                  - An integer repesenting the program counter of the context
                    (if it is suspended)
                  - A string representing the reason why the context is
                    suspended
                  - A dictionary of properties stating why the suspended state
                    properties (see runcontrol.STATE_*)
    """
    def callGetState(context, condition, val):
        """Asynchronous request to get the context state.

        :param context: The RunControlContext to get state for.
        :param condition: A threading.Condition the caller is pending on.
                          Caller is released from waiting through a
                          Condition.notify() call.
        :param val: A TcfValue handling the request returned value.

        :returns: **None**, always.
        """
        class DoneGetState(runcontrol.DoneGetState):
            """Client call back class for RunControlContext.getState()."""
            def doneGetState(self, token, error, suspended, pc, reason,
                             params):
                """Called when RunControlContext.getState() command execution
                is complete.

                :param token: pending command handle.
                :param error: command execution error or None.
                :param suspended: true if the context is suspended
                :param pc: program counter of the context (if suspended).
                :param reason: suspend reason (if suspended), see REASON_*.
                :param params: additional target specific data about context
                               state, see STATE_*.
                """
                if error:
                    protocol.log("Error from runcontrol.getState()", error)
                else:
                    val.setValue((suspended, pc, reason, params))
                with condition:
                    condition.notify()

        # start the process itself

        context.getState(DoneGetState())

    # create a condition to wait on, and a value to get the children ids

    lock = threading.Condition()
    value = TcfValue()

    with lock:
        # TCF requests must be called by the dispatch thread, wait for a
        # maximum of 10 seconds
        protocol.invokeLater(callGetState, context=context, condition=lock,
                             val=value)
        lock.wait(10)

    return (value.getValue())


# Some TCF initialisation


def subscribe(svc, streamType, listener):
    """Subscribe to a streams channel.

    :param svc: The TCF streams proxy to subscribe against.
    :param streamType: Type of the stream to register against. For now,
                       I only know of 'Terminals', 'Processes' and
                       'ProcessesV1' types.
    :param listener: The listener to subscribe to *svc*.
    :type listener: tcf.services.streams.StreamsListener

    :returns: **None**, always.
    """
    def callSubscribe(service, streamType, listener, condition):
        """Asynchronous request to subscribe a listener to streams service.

        :param service: The streams service to subscribe listener against.
        :param streamType: Type of the stream to register against. For now,
                           I only know of 'Terminals', 'Processes' and
                           'ProcessesV1' types.
        :param listener: The listener to subscribe to *service*.
        :type listener: tcf.services.streams.StreamsListener
        :param condition: A threading.Condition the caller is pending on.
                          Caller is released from waiting through a
                          Condition.notify() call.

        :returns: **None**, always.
        """
        class DoneSubscribe(streams.DoneSubscribe):
            """Call back interface for StreamsService.subscribe() command."""
            def doneSubscribe(self, token, error):
                """Called when stream subscription is done.

                :param token: pending command handle
                :param error: error description if operation failed, **None**
                              if succeeded.
                """
                if error:
                    protocol.log("Error from streams.subscribe()", error)
                with condition:
                    condition.notify()

        # start the process itself

        service.subscribe(streamType, listener, DoneSubscribe())

    # create a condition to wait on

    lock = threading.Condition()

    with lock:
        # TCF requests must be called by the dispatche thread, wait for a
        # maximum of 10 seconds
        protocol.invokeLater(callSubscribe, service=svc, streamType=streamType,
                             listener=listener, condition=lock)
        lock.wait(10)

# --------------------------------------------------------------------------- #

# TCF initialisation

protocol.startEventQueue()
protocol.setLogger(TcfProtocolLogger())
atexit.register(protocol.getEventQueue().shutdown)

try:
    c = tcf.connect("TCP:127.0.0.1:1534")
except Exception as e:
    protocol.log(e)
    sys.exit()

# If there is a streams service, listen to it

streamsSvc = getService(c, streams.NAME)
if streamsSvc:
    subscribe(streamsSvc, 'ProcessesV1', StreamsListener(streamsSvc))

p = start(c, '/pdi_reader/ls', '-l', '-a')

# this part is a bit tricky, the runcontrol contexts which accept a resume
# command are the contexts which have a state. Recursively try to find the
# runcontrol context which has a state, and resume it.

rcSvc = getService(c, runcontrol.NAME)

if rcSvc is None:
    print('No runcontrol service. Exiting ...')
    sys.exit()

context = getContext(rcSvc, p.getID())
print('Runcontrol context is a container: ' + str(context.isContainer()))

while context and not context.hasState():
    children = getChildren(rcSvc, context.getID())
    for child in children:
        context = getContext(rcSvc, child)
        if context and context.hasState():
            break

if context is None:
    print('No runcontrol context to resume. Exiting ...')
    sys.exit()

# get the state of this context. State is a tuple of
# (suspended, pc, reason, params)

ctxState = state(context)
print('Context state : ' + str(ctxState))

while ctxState and ctxState[0]:
    resume(context)
    # calling resume may end the context ... catch exceptions

    try:
        ctxState = state(context)
        if ctxState:
            print('Context state : ' + str(ctxState))
    except:
        pass

# Let the async calls the time to end ...

time.sleep(2)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
from chipscopy.dm import jtag, memory, request, NodeAutoUpgrader
from chipscopy.tcf.services import xicom
from ..jtagdevice import JtagDevice, JtagCable
from ..mem import MemoryNode


def find_pmc_tap(hws):
    view = hws.get_view(jtag)
    for node in view.get_all():
        if JtagDevice.is_compatible(node) and (
            node.arch_name and (node.arch_name == "everest" or node.arch_name == "versal")
        ):
            return view.get_node(node.ctx, JtagDevice)
    raise (Exception("Could not find pmc_tap"))


def find_jtag_device(hws):
    view = hws.get_view(jtag)
    for node in view.get_all():
        if JtagDevice.is_compatible(node):
            return view.get_node(node.ctx, JtagDevice)
    raise (Exception("Could not find pmc_tap"))


class PmcTapUtil(JtagDevice):
    def status(self, reg_names: List[str] = ("jtag_status",), done: request.DoneCallback = None):
        """
        Gets status values of registers available on the pmc_tap.
        :param reg_names: Names of desired registers
        :param done: DoneRequest callback
        :return: Token of request
        """
        proc = self.manager.channel.getRemoteService("XicomEverest")
        done = request._make_callback(done)
        node = self
        need_definitions = False

        if "status" not in self.props:
            self.props["status"] = {}
        for name in reg_names:
            if self.get_def_name(name) not in self.props["status"].keys():
                need_definitions = True
                break

        class DoneCmd(xicom.DoneXicom):
            def doneHW(self, token, error, props):
                print("Props:", props)
                if not error:
                    node["status"].update(props)
                    _print_regs(node["status"])
                node.remove_pending(token)  # done state change
                done.done_request(token, error, props)

        # add pending to indicate that this node is being changed
        return self.add_pending(proc.get_status(self.ctx, need_definitions, DoneCmd()))


def _print_regs(regs):
    for reg_name, reg_value in regs.items():
        if isinstance(reg_value, bytearray):
            definition = reg_name + "_definition"
            if definition in regs.keys():
                print(f"\n{reg_name.upper()}")
                print("-" * 60)
                _print_defs(regs[definition], reg_value)
            else:
                reg_value = bytearray(reg_value)
                reg_value.reverse()
                print(f"{reg_name.upper()}: {reg_value.hex()}")


def _to_bits(data):
    return "".join(str((byte >> b) & 1) for byte in data for b in range(8))


def _get_bits(bits, select):
    first, _, last = select.partition(":")
    first = int(first)
    last = int(last) if last else first
    if last < first:
        return bits[first : last - 1 : -1]
    return bits[first : last + 1]


FIELD_WIDTH = 20
BIT_WIDTH = 8


def _print_defs(defs, data):
    bits = _to_bits(data)
    print("{} {} {}".format("Bit".ljust(BIT_WIDTH, " "), "Field".rjust(FIELD_WIDTH, " "), "Value"))
    for bit_range, bit_name in zip(defs["bit_range"], defs["bit_name"]):
        print(
            "{} {} {}".format(
                bit_range.ljust(BIT_WIDTH, " "),
                bit_name.rjust(FIELD_WIDTH, " "),
                _get_bits(bits, bit_range),
            )
        )


class JtagNodeAutoUpgrader(NodeAutoUpgrader):
    def __call__(self, node, **kwargs):
        if JtagCable.is_compatible(node):
            node = node.manager.upgrade_node(node, JtagCable)
        elif JtagDevice.is_compatible(node):
            node = node.manager.upgrade_node(node, JtagDevice)
        node["node_cls"] = node.__class__
        return node


class MemoryNodeAutoUpgrader(NodeAutoUpgrader):
    def __call__(self, node, **kwargs):
        if MemoryNode.is_compatible(node):
            node = node.manager.upgrade_node(node, MemoryNode)
        node["node_cls"] = node.__class__
        return node


def init_config():
    if not jtag.JtagManager.node_auto_upgrader:
        jtag.JtagManager.node_auto_upgrader = JtagNodeAutoUpgrader()
    if not memory.MemoryManager.node_auto_upgrader:
        memory.MemoryManager.node_auto_upgrader = MemoryNodeAutoUpgrader()


def test():
    from chipscopy import client

    # from chipscope.utils import logger
    # logger.enable_domain('view_info')
    # logger.change_log_level("debug")
    hws = client.connect("xcordevl154")
    hws.jtag_target()
    t = hws.jtag_target(index=0)
    return hws, hws.jtag_target(index=1, parent=t)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from enum import Enum
from typing import Union, Optional

from typing_extensions import Final, Literal

from rich.text import Text
from rich.tree import Tree
from rich.table import Table
from rich.style import Style
from rich.console import Console
from rich.progress import Progress, TaskID, TextColumn, BarColumn


class Printer:
    def __init__(self):
        self.console = Console()

    @property
    def quiet(self):
        return self.console.quiet

    @quiet.setter
    def quiet(self, value):
        self.console.quiet = value

    def __call__(self, *args, level: Optional[Literal["info", "warning"]] = None):
        if isinstance(args[0], Table) or isinstance(args[0], Tree):
            self.console.print(args[0])
        elif len(args) == 1 and isinstance(args[0], str):
            if level is not None:
                self._print_msg(args[0], level=level)
            else:
                self.console.print(args[0])
        else:
            self.console.print(*args)

    def _print_msg(self, msg, *, level: Literal["info", "warning"]):
        custom_styles: Final[dict] = {
            "info": Style(bold=False),
            "warning": Style(color="yellow", bold=False, italic=True),
        }

        if level == "info":
            full_msg = Text("\n--> INFO: ", style=Style(bold=True))
        elif level == "warning":
            full_msg = Text("\n--> WARNING: ", style=Style(color="yellow", bold=True))
        else:
            raise ValueError(f"Unknown print level '{level}'")

        default_style = custom_styles[level]

        msgs = msg.split("\n")

        full_msg.append(msgs[0], style=default_style)

        if len(msgs) > 1:
            for msg in msgs[1:]:
                original_msg = msg
                newline_chars = "\n"

                if msg.startswith("\n"):
                    match = re.match("^(\\n+)(.*)", msg)
                    newline_chars, original_msg = match.groups()

                # This is for aligning the current line with the first line of the message
                msg = f"{newline_chars}{' ' * 4}{original_msg}"
                full_msg.append(msg, style=default_style)

        self.console.print(full_msg)


printer = Printer()


class PercentProgressBar:
    class Status(Enum):
        DONE = "[bold green]Done[/]"
        ABORTED = "[bold red]Aborted[/]"
        STARTING = "[bold yellow]Starting[/]"
        IN_PROGRESS = "[bold blue]In Progress[/]"

    def __init__(self):
        self.progress_meter = Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TextColumn("{task.percentage:>3.0f}%", style="bold bright_magenta"),
            TextColumn("{task.fields[status]}"),
            console=printer.console,
            auto_refresh=False,
        )
        self.task_id: TaskID = None

    def add_task(
        self,
        *,
        total: int = 100,
        status: Union[str, Status],
        visible: bool = True,
        description: str,
    ):
        self.task_id: TaskID = self.progress_meter.add_task(
            description,
            total=total,
            status=status if isinstance(status, str) else status.value,
            visible=visible,
        )

        # Start the progress meter everytime we add a task.
        # If its already running, rich knows nothing needs to be done
        self.progress_meter.start()

        return self.task_id

    def update(
        self,
        *,
        status: Union[str, Status],
        task_id: TaskID = None,
        completed: int = None,
    ):
        if task_id is None:
            if self.task_id is None:
                return
            task_id = self.task_id

        kwargs = {"refresh": True, "status": status if isinstance(status, str) else status.value}
        if completed is not None:
            kwargs["completed"] = completed

        self.progress_meter.update(task_id, **kwargs)

        if kwargs["status"] in {
            PercentProgressBar.Status.DONE.value,
            PercentProgressBar.Status.ABORTED.value,
        }:
            if self.progress_meter.finished:
                self.progress_meter.stop()


if __name__ == "__main__":
    printer("\n\nJust your normal print. Nothing special to see here...")

    printer("Hello there, this is an INFO message", level="info")
    printer("Hello there\nThis is an INFO message\nspread over multiple lines", level="info")
    printer(
        "Hello there\nThis is an INFO message\n"
        "spread over multiple lines\n\nwith blank lines in between",
        level="info",
    )

    printer("Hello there, this is a WARNING message", level="warning")
    printer("Hello there\nThis is an WARNING message\nspread over multiple lines", level="warning")
    printer(
        "Hello there\nThis is a WARNING message\n"
        "spread over multiple lines\n\nwith blank lines in between",
        level="warning",
    )

    printer("\n")

    import threading
    import time

    def func1(task_id, sleep_time):
        x = 0
        while True:
            time.sleep(sleep_time)
            progress_meter.update(task_id=task_id, status="In Progress", completed=x)
            x += 1

            if x > 100:
                break

        progress_meter.update(task_id=task_id, status="Done")

    progress_meter = PercentProgressBar()
    task_id1 = progress_meter.add_task(
        description=f"Thread 1 progress ",
        status="Starting",
    )

    t1 = threading.Thread(target=func1, args=(task_id1, 0.1))
    t1.start()

    task_id2 = progress_meter.add_task(
        description=f"Thread 2 progress ",
        status="Starting",
    )

    t2 = threading.Thread(target=func1, args=(task_id2, 0.2))

    t2.start()

    t1.join()
    t2.join()

    printer.quiet = True

    printer("You shouldn't see this message, because the printer is turned off")



# *****************************************************************************
# * Copyright (c) 2011, 2013, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

import threading

from . import protocol, channel
from .services import locator

_channels = []
_listeners = []
_transports = {}
_lock = threading.RLock()


class TransportProvider(object):
    """
    TransportProvider represents communication protocol that can be used to
    open TCF communication channels.
    Examples of transports are: TCP/IP, RS-232, USB.

    Client can implement this interface if they want to provide support for a
    transport that is not supported directly by the framework.
    """

    def getName(self):
        """
        Return transport name. Same as used as peer attribute,
        @see IPeer.ATTR_TRANSPORT_NAME
        @return transport name.
        """
        raise NotImplementedError("Abstract method")

    def openChannel(self, peer):
        """
        Open channel to communicate with this peer using this transport.
        Note: the channel can be not fully open yet when this method returns.
        It's state can be IChannel.STATE_OPENING.
        Protocol.Listener will be called when the channel will be opened or
        closed.
        @param peer - a IPeer object that describes remote end-point of the
                      channel.
        @return TCF communication channel.
        """
        raise NotImplementedError("Abstract method")


def addTransportProvider(transport):
    name = transport.getName()
    assert name
    with _lock:
        if _transports.get(name):
            raise Exception("Already registered: " + name)
        _transports[name] = transport


def removeTransportProvider(transport):
    name = transport.getName()
    assert name
    with _lock:
        if _transports.get(name) == transport:
            del _transports[name]


def openChannel(peer):
    name = peer.getTransportName()
    if not name:
        raise Exception("No transport name")
    with _lock:
        provider = _transports.get(name)
        if not provider:
            raise Exception("Unknown transport name: " + name)
    return provider.openChannel(peer)


def channelOpened(channel):
    assert channel not in _channels
    _channels.append(channel)
    for l in tuple(_listeners):
        try:
            l.onChannelOpen(channel)
        except Exception as x:
            import sys
            x.tb = sys.exc_info()[2]
            protocol.log("Exception in channel listener", x)


def channelClosed(channel, error):
    assert channel in _channels
    _channels.remove(channel)


def getOpenChannels():
    return _channels[:]


def addChannelOpenListener(listener):
    assert listener
    _listeners.append(listener)


def removeChannelOpenListener(listener):
    try:
        _listeners.remove(listener)
    except ValueError:
        pass  # ignore


class TCPTransportProvider(TransportProvider):
    def getName(self):
        return "TCP"

    def openChannel(self, p):
        assert self.getName() == p.getTransportName()
        from . import peer
        attrs = p.getAttributes()
        host = attrs.get(peer.ATTR_IP_HOST)
        port = attrs.get(peer.ATTR_IP_PORT)
        if not host:
            raise RuntimeError("No host name")
        from .channel.ChannelTCP import ChannelTCP
        return ChannelTCP(p, host, _parsePort(port))


class LoopbackTransportProvider(TransportProvider):
    def getName(self):
        return "loopback"

    def openChannel(self, p):
        assert self.getName() == p.getTransportName()
        from .channel.LoopbackChannel import LoopbackChannel
        return LoopbackChannel(p)


def _parsePort(port):
    if not port:
        raise Exception("No port number")
    try:
        return int(port)
    except Exception:
        raise RuntimeError("Invalid value of \"Port\" attribute. Must be " +
                           "decimal number.")


def sendEvent(service_name, event_name, data):
    """
    Transmit TCF event message.
    The message is sent to all open communication channels - broadcasted.

    This is internal API, TCF clients should use protocol.sendEvent().
    """
    for c in _channels:
        # Skip channels that are executing "redirect" command - STATE_OPENING
        if c.getState() == channel.STATE_OPEN:
            s = c.get_local_service(service_name)
            if s:
                c.sendEvent(s, event_name, data)


def sync(done):
    """
    Call back after TCF messages sent by this host up to this moment are
    delivered to their intended targets. This method is intended for
    synchronization of messages across multiple channels.

    Note: Cross channel synchronization can reduce performance and throughput.
    Most clients don't need cross channel synchronization and should not call
    this method.

    @param done will be executed by dispatch thread after communication
    messages are delivered to corresponding targets.

    This is internal API, TCF clients should use protocol.sync().
    """
    tokenSet = set()

    class DoneSync(locator.DoneSync):
        def doneSync(self, token):
            assert tokenSet.contains(token)
            tokenSet.remove(token)
            if len(tokenSet) == 0:
                done()
    done_sync = DoneSync()
    for c in _channels:
        if c.getState() == channel.STATE_OPEN:
            s = c.getRemoteService(locator.NAME)
            if s:
                tokenSet.append(s.sync(done_sync))
    if len(tokenSet) == 0:
        protocol.invokeLater(done)


# initialize TCP and Loopback transports
addTransportProvider(TCPTransportProvider())
addTransportProvider(LoopbackTransportProvider())



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import enum
import json
from abc import abstractmethod
from collections import defaultdict
from dataclasses import dataclass, asdict
from datetime import datetime
from io import TextIOBase
from itertools import islice
from pprint import pformat
from typing import Generator, Dict, List, Union, Optional
from chipscopy.api.ila import ILABitRange, ILAProbeRadix
import chipscopy
from chipscopy.shared.ila_util import bin_reversed_to_hex_values
from chipscopy.utils import Enum2StrEncoder


@dataclass
class ILAWaveformProbe:
    """Probe location in a data sample."""

    name: str
    """Name of probe."""
    map: str
    """Location string"""
    map_range: List[ILABitRange]
    """List of bit ranges. See :class:`~chipscopy.api.ila.ILABitRange`"""
    is_bus: bool
    """True for bus probes"""
    bus_left_index: Optional[int] = None
    """ Bus left index. E.g. 5 in probe ``counter[5:0]``"""
    bus_right_index: Optional[int] = None
    """Bus right index. E.g. 0 in probe ``counter[5:0]``"""
    display_radix: ILAProbeRadix = ILAProbeRadix.HEX
    """Display radix, when exporting waveform data. Default is ILAProbeRadix.HEX"""
    enum_def: Optional[enum.EnumMeta] = None
    """Enum class defining {name:int} enum values, for this probe."""

    def length(self) -> int:
        return sum(mr.length for mr in self.map_range)

    def bus_range_str(self) -> str:
        if not self.is_bus:
            res = ""
        elif self.bus_left_index == self.bus_right_index:
            res = f"[{self.bus_left_index}]"
        else:
            res = f"[{self.bus_left_index}:{self.bus_right_index}]"
        return res


@dataclass
class ILAWaveform:
    """Waveform data, with data probe information."""

    width: int
    """Sample bit width."""
    sample_count: int
    """Number of data samples."""
    trigger_position: List[int]
    """Trigger position index, for each data window."""
    window_size: int
    """Number of samples in a window."""
    probes: Dict[str, ILAWaveformProbe]
    """Dict of {probe name, waveform probe}   See :class:`ILAWaveformProbe`"""
    data: bytearray
    """
    Waveform data.

    Samples are aligned on byte boundary. 

    This formula can be used to read a bit from the data:
    ::

        bytes_per_sample = len(data) // sample_count

        def get_bit_value(data: bytearray, bytes_per_sample: int,
                          sample_index: int, data_bit_index: int) -> bool:
            byte_value = data[sample_index * bytes_per_sample + data_bit_index // 8]
            mask = 1 << (data_bit_index & 0x7)
            return (byte_value & mask) != 0

    """

    def get_window_count(self) -> int:
        return len(self.trigger_position)

    def __str__(self) -> str:
        items = {key: val for key, val in self.__dict__.items() if key != "data"}
        return pformat(items, 2)

    def __repr__(self) -> str:
        data_size = len(self.data) if self.data else 0
        json_dict = {
            "width": self.width,
            "sample_count": self.sample_count,
            "trigger_position": self.trigger_position,
            "window_size": self.window_size,
            "probes": {key: asdict(val) for key, val in self.probes.items()},
            "data size": hex(data_size),
        }
        ret = json.dumps(json_dict, cls=Enum2StrEncoder, indent=4)
        return ret


def tcf_get_waveform_data(tcf_node) -> {}:
    tcf_props = tcf_node.get_property_group(["data"])
    props = {
        "width": tcf_props["trace_width"],
        "sample_count": tcf_props["trace_sample_count"],
        "trigger_position": tcf_props["trace_trigger_position"],
        "window_size": tcf_props["trace_window_size"],
        "data": tcf_props["trace_data"],
    }
    return props


class WaveformWriter(object):
    def __init__(self, file_handle: Union[TextIOBase, None], probes: [ILAWaveformProbe]):
        self._file_handle = file_handle
        self._probes = probes
        self._probe_count = len(probes)
        self._probe_names = self.make_probe_names()
        self._probe_widths = [p.length() for p in probes]

    def handle(self) -> TextIOBase:
        return self._file_handle

    def write(self, msg: str):
        self._file_handle.write(msg)

    @abstractmethod
    def make_probe_names(self) -> [str]:
        return [p.name for p in self._probes]

    @abstractmethod
    def init(self):
        pass

    @abstractmethod
    def write_sample(
        self,
        sample_position: int,
        window_index: int,
        sample_in_window_index: int,
        is_trigger: bool,
        probe_binary_reversed_values: [str],
        is_last_sample: bool,
    ) -> None:
        """

        Args:
            sample_position (int):
            window_index (int):
            sample_in_window_index (int):
            is_trigger (bool):
            probe_binary_reversed_values ([str]): binary string values each start lsb at position zero.
            is_last_sample(bool): Last sample in waveform.

        Returns:

        """
        pass


class WaveformWriterCSV(WaveformWriter):
    def __init__(self, file_handle: TextIOBase, probes: [ILAWaveformProbe]):
        WaveformWriter.__init__(self, file_handle, probes)

    def init(self):
        self.write("Sample in Buffer,Sample in Window,TRIGGER,")
        self.write(",".join(self._probe_names))
        self.write("\nRadix - UNSIGNED,UNSIGNED,UNSIGNED,")
        # currently HEX is the only supported radix
        self.write(",".join(["HEX"] * self._probe_count))
        self.write("\n")

    def make_probe_names(self) -> [str]:
        return [p.name + p.bus_range_str() for p in self._probes]

    def write_sample(
        self,
        sample_position: int,
        window_index: int,
        sample_in_window_index: int,
        is_trigger: bool,
        probe_binary_reversed_values: [str],
        is_last_sample: bool,
    ) -> None:
        hex_values = bin_reversed_to_hex_values(probe_binary_reversed_values)
        hex_values_str = ",".join(hex_values)
        trig_mark = "1" if is_trigger else "0"
        self.write(f"{sample_position},{sample_in_window_index},{trig_mark},{hex_values_str}\n")


class WaveformWriterToDict(WaveformWriter):
    def __init__(
        self, probes: [ILAWaveformProbe], include_trigger: bool, include_sample_info: bool
    ):
        WaveformWriter.__init__(self, None, probes)
        self._include_trigger = include_trigger
        self._include_sample_info = include_sample_info
        self._result = defaultdict(list)

    def init(self):
        return

    def get_result(self) -> Dict[str, List[int]]:
        return self._result

    def write_sample(
        self,
        sample_position: int,
        window_index: int,
        sample_in_window_index: int,
        is_trigger: bool,
        probe_binary_reversed_values: [str],
        is_last_sample: bool,
    ) -> None:
        if self._include_trigger:
            self._result["__TRIGGER"].append(1 if is_trigger else 0)
        if self._include_sample_info:
            self._result["__SAMPLE_INDEX"].append(sample_position)
            self._result["__WINDOW_INDEX"].append(window_index)
            self._result["__WINDOW_SAMPLE_INDEX"].append(sample_in_window_index)

        int_values = [int(r_val[::-1], 2) for r_val in probe_binary_reversed_values]
        for probe_name, val in zip(self._probe_names, int_values):
            self._result[probe_name].append(val)


class WaveformWriterVCD(WaveformWriter):
    """Value Change Dump format. See Wikipedia and IEEE Std 1364-2001."""

    def __init__(self, file_handle: TextIOBase, probes: [ILAWaveformProbe]):
        WaveformWriter.__init__(self, file_handle, probes)
        self._values = None
        self._vars = None
        self._trigger_var = None
        self._window_var = None
        self._prev_sample_is_trigger = None
        self._prev_window_index = None

    @staticmethod
    def _generate_vars() -> Generator[str, None, None]:
        """Supports up to 94 + 94*94 + 94*94*94 = 839,514 probes (or variables)."""
        for x in range(ord("!"), ord("~")):
            yield chr(x)
        for xx in range(ord("!"), ord("~")):
            for yy in range(ord("!"), ord("~")):
                yield chr(xx) + chr(yy)
        for xxx in range(ord("!"), ord("~")):
            for yyy in range(ord("!"), ord("~")):
                for zzz in range(ord("!"), ord("~")):
                    yield chr(xxx) + chr(yyy) + chr(zzz)

    def _write_variable_definitions(self):
        self._values = [None] * self._probe_count
        self._trigger_var, self._window_var, *self._vars = islice(
            WaveformWriterVCD._generate_vars(), self._probe_count + 2
        )

        self._trigger_var = self._trigger_var
        for width, var, probe_name in zip(self._probe_widths, self._vars, self._probe_names):
            self.write(f"$var reg {width} {var} {probe_name} $end\n")
        self.write(
            f"$var reg 1 {self._trigger_var} _TRIGGER $end\n"
            f"$var reg 1 {self._window_var} _WINDOW $end\n"
        )

    def make_probe_names(self) -> [str]:
        return [p.name + " " + p.bus_range_str() for p in self._probes]

    def init(self):
        now = "{:%Y-%m-%d %H:%M:%S}".format(datetime.now())
        # adjust for trigger position
        hdr_1 = f"$date\n        {now}\n$end\n"

        hdr_2 = (
            "$version\n"
            + f"        ChipScoPy Version {chipscopy.__version__}\n"
            + "$end\n"
            + "$timescale\n        1ps\n$end\n"
            + "$scope module dut $end\n"
        )

        hdr_3 = "$upscope $end\n$enddefinitions $end\n"

        self.write(hdr_1 + hdr_2)
        self._write_variable_definitions()
        self.write(hdr_3)

    def write_sample(
        self,
        sample_position: int,
        window_index,
        sample_in_window_index,
        is_trigger,
        probe_binary_reversed_values: [str],
        is_last_sample,
    ):
        time_written = False

        def write_value(var: str, reversed_value: str) -> None:
            nonlocal time_written
            if not time_written:
                self.write(f"#{sample_position}\n")
                time_written = True

            if len(reversed_value) == 1:
                self.write(f"{reversed_value}{var}\n")
                return

            # Remove leading zeroes
            msb_idx = reversed_value.rfind("1")
            if msb_idx < 0:
                msb_idx = 0
            value = reversed_value[msb_idx::-1]
            self.write(f"b{value} {var}\n")

        # Write time, for last sample, even if no changes.
        if is_last_sample:
            self.write(f"#{sample_position}\n")
            time_written = True

        # Trigger value
        if is_trigger != self._prev_sample_is_trigger:
            write_value(self._trigger_var, "1" if is_trigger else "0")
            self._prev_sample_is_trigger = is_trigger

        # Window marker.
        if window_index != self._prev_window_index:
            write_value(self._window_var, "1" if window_index % 2 else "0")
            self._prev_window_index = window_index

        # Regular values
        for idx, new_val in enumerate(probe_binary_reversed_values):
            if new_val == self._values[idx]:
                continue

            self._values[idx] = new_val
            write_value(self._vars[idx], new_val)


def export_waveform_to_stream(
    waveform: ILAWaveform,
    export_format: str,
    stream_handle: TextIOBase,
    probe_names: [str],
    start_window_idx: int,
    window_count: Optional[int],
    start_sample_idx: int,
    sample_count: Optional[int],
) -> None:
    """Arguments documented in calling API function"""
    if probe_names:
        probes = [waveform.probes.get(p_name, None) for p_name in probe_names]
        if not all(probes):
            bad_names = set(probe_names) - set(waveform.probes.keys())
            raise KeyError(f"export_waveform() called with non-existent probe_name:\n  {bad_names}")
    else:
        probes = waveform.probes.values()

    if export_format.upper() == "CSV":
        waveform_writer = WaveformWriterCSV(stream_handle, probes)
    elif export_format.upper() == "VCD":
        waveform_writer = WaveformWriterVCD(stream_handle, probes)
    else:
        raise ValueError(f'export_waveform() called with non-supported format "{export_format}"')

    _export_waveform(
        waveform,
        waveform_writer,
        probes,
        start_window_idx,
        window_count,
        start_sample_idx,
        sample_count,
        "export_waveform()",
    )


def get_waveform_data_values(
    waveform: ILAWaveform,
    probe_names: [str],
    start_window_idx: int,
    window_count: Optional[int],
    start_sample_idx: int,
    sample_count: Optional[int],
    include_trigger: bool,
    include_sample_info: bool,
) -> Dict[str, List[int]]:
    """Arguments documented in calling API function"""
    if probe_names:
        probes = [waveform.probes.get(p_name, None) for p_name in probe_names]
        if not all(probes):
            bad_names = set(probe_names) - set(waveform.probes.keys())
            raise KeyError(
                f"get_waveform_probe_data() called with non-existent probe name(s):\n  {bad_names}"
            )
    else:
        probes = waveform.probes.values()

    waveform_writer = WaveformWriterToDict(probes, include_trigger, include_sample_info)
    _export_waveform(
        waveform,
        waveform_writer,
        probes,
        start_window_idx,
        window_count,
        start_sample_idx,
        sample_count,
        "get_waveform_data()",
    )

    return waveform_writer.get_result()


def _export_waveform(
    waveform: ILAWaveform,
    writer: WaveformWriter,
    probes: [ILAWaveformProbe],
    start_window_idx: int,
    window_count: Optional[int],
    start_sample_idx: int,
    sample_count: Optional[int],
    calling_function: str,
) -> None:
    def get_sample_binary_values(sample_data: memoryview) -> [str]:
        sample_int_value = int.from_bytes(sample_data, byteorder="little", signed=False)
        reversed_bin_value = f"{sample_int_value:0{waveform.width}b}"[::-1]
        values = []
        for probe in probes:
            # Each probe value can be made up of multiple bit range slices.
            slices = [reversed_bin_value[br.index : br.index + br.length] for br in probe.map_range]
            p_val = "".join(slices)
            values.append(p_val)
        return values

    if not window_count:
        window_count = waveform.get_window_count()
    if not sample_count:
        sample_count = waveform.window_size

    w_size = waveform.window_size
    max_window_count = (waveform.sample_count + w_size - 1) // w_size
    if start_window_idx < 0 or start_window_idx >= max_window_count:
        raise ValueError(
            f'{calling_function} function argument start_window="{start_window_idx}" '
            f"must be in the range [0-{max_window_count - 1}]"
        )
    if start_sample_idx < 0 or start_sample_idx >= w_size:
        raise ValueError(
            f'{calling_function} function argument "start_sample="{start_sample_idx}" '
            f"must be in the range [0-{w_size - 1}]"
        )
    if sample_count < 1 or sample_count > w_size - start_sample_idx:
        raise ValueError(
            f'{calling_function} function argument "sample_count="{sample_count}" '
            f"must be in the range [1-{w_size - start_sample_idx}], "
            f'since start_sample_idx="{start_sample_idx}".'
        )
    if window_count < 1 or window_count > max_window_count - start_window_idx:
        raise ValueError(
            f'{calling_function} function argument "window_count="{window_count}" '
            f"must be in the range [1-{max_window_count - start_window_idx}], "
            f'since start_window_idx="{start_window_idx}".'
        )

    sample_byte_count = (waveform.width + 7) // 8
    raw_samples = memoryview(waveform.data)

    writer.init()

    for window_idx in range(start_window_idx, start_window_idx + window_count):
        window_start_sample_idx = window_idx * w_size
        trigger_sample_idx = waveform.trigger_position[window_idx]
        for sample_idx in range(
            window_start_sample_idx + start_sample_idx,
            window_start_sample_idx + start_sample_idx + sample_count,
        ):
            if sample_idx >= waveform.sample_count:
                # last window is not full.
                break
            raw_sample_idx = sample_idx * sample_byte_count
            raw_sample_idx_end = raw_sample_idx + sample_byte_count
            sample_idx_in_window = sample_idx - window_start_sample_idx
            bin_values = get_sample_binary_values(raw_samples[raw_sample_idx:raw_sample_idx_end])
            writer.write_sample(
                sample_idx,
                window_idx,
                sample_idx_in_window,
                sample_idx_in_window == trigger_sample_idx,
                bin_values,
                sample_idx + 1 == waveform.sample_count,
            )



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from collections import deque

import click
from more_itertools import one

import chipscopy
from chipscopy.api.device.device_scanner import DeviceScanner
from chipscopy import create_session, report_versions, report_devices


class NodeVisitorBase:
    """
    In ChipScoPy, nodes are organized as a tree under a view.
    Example hw_server views are: jtag, memory, debugcore
    Example cs_server views are: chipscope

    Hierarchy:
    server -> view -> node tree

    NodeVisitorBase is a base class that makes visiting nodes easy. Just make a class with the
    NodeVisitorBase as a parent, and define functions with visit_<cls_name> for the node types
    you want to visit. <cls_name> represents the node class type like "JtagDevice", etc.
    """

    def __init__(self, view):
        self._view = view

    def walk(self, node_list=None):
        """
        Recursively walk all nodes in view, including this node
        """
        if node_list is None:
            nodes_to_walk = list(self._view.get_children())
        else:
            nodes_to_walk = list(node_list)

        if len(nodes_to_walk) > 0:
            last_node = nodes_to_walk[-1]
        else:
            last_node = True
        if nodes_to_walk:
            for current_node in nodes_to_walk:
                self._visit(current_node, current_node == last_node)

    def _visit(self, node, is_last_sibling):
        cls = node.props.get("node_cls")
        # Example: <class \'chipscopy.client.jtagdevice.JtagCable\'>
        match = re.fullmatch("^<class \\'(.+)\\'>$", str(cls))
        full_class_name = match.group(1)
        class_name = full_class_name.split(".")[-1]
        method = "_visit_" + class_name
        visitor_method = getattr(self, method, self._visit_default)
        upgraded_node = self._view.get_node(node.ctx, cls)
        assert upgraded_node is not None
        return visitor_method(upgraded_node, is_last_sibling)

    def _visit_default(self, node, is_last_sibling):
        children = list(self._view.get_children(node))
        self.walk(children)

    @staticmethod
    def _print_node_path(node_name, parent_names):
        parent_names = deque(parent_names)
        while parent_names:
            name = parent_names.popleft()
            print(f"{name}/", end="")
        print(f"{node_name}")


# GLOBALS FOR COMMAND LINE MODULE
_hw_url: str
_cs_url: str


def display_banner():
    print(f"\n******** Xilinx ChipScoPy v{chipscopy.__version__}")
    print("  ****** Copyright 2021 Xilinx, Inc. All Rights Reserved.\n")


@click.group()
@click.version_option(version=chipscopy.__version__, message="%(version)s")
@click.option(
    "--hw_url",
    envvar="HW_SERVER_URL",
    default="localhost:3121",
    metavar="<url>",
    show_default=True,
    help="hardware server url",
)
@click.option(
    "--cs_url",
    envvar="CS_SERVER_URL",
    default="None",
    metavar="<url>",
    show_default=True,
    help="chipscope server url",
)
def _chipscopy(hw_url, cs_url):
    global _hw_url, _cs_url
    if len(hw_url) == 0 or hw_url.lower() == "none":
        _hw_url = None
    else:
        _hw_url = hw_url
    if len(cs_url) == 0 or cs_url.lower() == "none":
        _cs_url = None
    else:
        _cs_url = cs_url


@click.command()
@click.option("--devices", is_flag=True, help="Report all devices")
@click.option("--servers", is_flag=True, help="Report all server version information")
def report(devices, servers):
    session = create_session(hw_server_url=_hw_url, cs_server_url=_cs_url)
    if devices:
        report_devices(session)
    elif servers:
        report_versions(session)
    else:
        print("report command needs an option like --devices or --servers")


def _create_path(device_count, hier_name):
    return f"/device/{device_count}/{hier_name}"


@click.command()
@click.argument("file")
def info(file):
    session = create_session(hw_server_url=_hw_url, cs_server_url=_cs_url)
    device_list = session.devices
    device_count = 0
    for device in device_list:
        for (
            server_type,
            view_name,
            name,
            context,
            hier_name,
        ) in device.device_identification.get_target_tuples():
            path = _create_path(device_count, hier_name)
            if path == file or context == file:
                print("target      :", path)
                print("server_type :", server_type)
                print("view        :", view_name)
                print("name        :", name)
                print("context     :", context)
                print("family      :", device.device_identification["family"])
                print("dna         :", hex(device.device_identification["dna"]))
                print("cable       :", device.device_identification["cable_name"])
                print("jtag_index  :", device.device_identification["jtag_index"])
        device_count += 1


@click.command()
@click.option("-l", "--long", is_flag=True, help="long listing including device contexts")
# @click.option("--force-detection", is_flag=True, help="Force debug core detection prior to list")
def ls(long):
    session = create_session(hw_server_url=_hw_url, cs_server_url=_cs_url)
    device_list = session.devices
    device_count = 0
    for device in device_list:
        for (
            server_type,
            view_name,
            name,
            context,
            hier_name,
        ) in device.device_identification.get_target_tuples():
            path = _create_path(device_count, hier_name)
            if long:
                print(f"{path:60s} {context}")
            else:
                print(f"{path}")
        device_count += 1


@click.command()
@click.option("--dna", default=None, required=False, help="find device to program with dna")
@click.option("--cable", default=None, required=False, help="jtag cable name")
@click.option(
    "--position", default=None, required=False, help="jtag position in chain (starting index is 0)"
)
@click.argument("file")
def program(file, dna, cable, position):
    session = create_session(hw_server_url=_hw_url, cs_server_url=_cs_url)
    args = {}
    if dna:
        args["dna"] = dna
    if cable:
        args["cable"] = cable
    if position:
        args["jtag_index"] = position
    if len(args) > 0:
        device = session.devices.get(**args)
    else:
        device = one(session.devices)

    device_tracker = device.device_identification
    print(f"\n{'Cable':40s} {'Jtag_Index':>10s}   {'Family':15s} {'DNA':30s}")
    print(
        f"{device_tracker['cable_name']:40s} "
        f"{device_tracker['jtag_index']:10d}   "
        f"{device_tracker['family']:15s} "
        f"{device_tracker['dna']:<30d}"
    )
    device.program(file)
    print()


@click.command(name="json_devices")
def json_devices():
    # click command string above is required because of underscore in name
    session = create_session(hw_server_url=_hw_url, cs_server_url=_cs_url)
    scanner = DeviceScanner(session.hw_server, session.cs_server)
    print(scanner)


class TreePrinter(NodeVisitorBase):
    tree_glyphs_std = {"blank": "    ", "line": "│   ", "middle": "├── ", "last": "└── "}
    tree_glyphs_alt = {"blank": "    ", "line": "|   ", "middle": "|-- ", "last": "\\-- "}
    tree_glyphs_space = {"blank": "    ", "line": "    ", "middle": "    ", "last": "    "}
    tree_glyphs_none = {"blank": "", "line": "", "middle": "", "last": ""}
    tree_glyphs_selector = {
        "std": tree_glyphs_std,
        "alt": tree_glyphs_alt,
        "space": tree_glyphs_space,
        "none": tree_glyphs_none,
    }

    def __init__(self, session, view_name):
        self._tree_glyphs = TreePrinter.tree_glyphs_std
        self._show_context = True
        self._jtag_index = None
        self.session = session
        self._view_name = view_name
        self._indent_prefix = list()

        if view_name == "chipscope":
            view = self.session.cs_server.get_view(view_name)
        else:
            view = self.session.hw_server.get_view(view_name)
        super().__init__(view)

    def print(self, *, show_context=False, glyph_type="std"):
        self._tree_glyphs = TreePrinter.tree_glyphs_selector.get(glyph_type)
        self._jtag_index = None
        self._show_context = show_context
        print(self._view_name)
        self._indent_prefix = list()
        self.walk()

    # def _print_node_properties(self, node):
    #     props_dict = node.props
    #     for key, val in props_dict.items():
    #         if key != "regs":
    #             print(f"        {key:}: {val}")
    #     # Special printing for all regs - they need to be refreshed
    #     regs = node.props.get("regs")
    #     if regs:
    #         reg_names_to_refresh = tuple(regs.keys())
    #         node.update_regs(reg_names=reg_names_to_refresh, force=True, done=None)
    #         for reg_key, reg_val in regs.items():
    #             bytearray_data = reg_val.data
    #             int_value = int.from_bytes(bytearray_data, byteorder="little", signed=False)
    #             print(f"        regs.{reg_key:}: {hex(int_value)}")

    def _print_node(self, node, prefix):
        name = node.props.get("Name")
        cls = node.props.get("node_cls")
        # Example: <class \'chipscopy.client.jtagdevice.JtagCable\'>
        match = re.fullmatch("^<class \\'(.+)\\'>$", str(cls))
        full_class_name = match.group(1)
        class_name = full_class_name.split(".")[-1]
        node_str = f"{prefix}{name} ({class_name})"
        if self._show_context:
            indent_character_count = 50 - len(node_str)
            if indent_character_count < 0:
                indent_character_count = 2
            indent = " " * indent_character_count
            print(f"{node_str}{indent}{node.ctx}")
        else:
            print(f"{node_str}")

    def _visit_JtagCable(self, node, is_last_sibling):
        # Tracking JtagCable and JtagDevice so we can keep running jtag_index per device in the chain
        self._jtag_index = 0
        self._visit_default(node, is_last_sibling)

    def _visit_JtagDevice(self, node, is_last_sibling):
        # Tracking JtagCable and JtagDevice so we can keep running jtag_index per device in the chain
        self._jtag_index += 1
        self._visit_default(node, is_last_sibling)

    def _visit_default(self, node, is_last_sibling):
        if is_last_sibling:
            glyph = self._tree_glyphs["last"]
        else:
            glyph = self._tree_glyphs["middle"]
        self._indent_prefix.append(glyph)
        prefix = "".join([i for i in self._indent_prefix])
        self._print_node(node, prefix)
        children = list(self._view.get_children(node))
        # Need to update prefix glyphs before moving to next nodes
        for idx, item in enumerate(self._indent_prefix):
            if item == self._tree_glyphs["last"]:
                self._indent_prefix[idx] = self._tree_glyphs["blank"]
            elif item == self._tree_glyphs["middle"]:
                self._indent_prefix[idx] = self._tree_glyphs["line"]
        self.walk(children)
        self._indent_prefix.pop()


@click.command()
@click.option(
    "--view", default="jtag", required=False, help="Use the specified view for walking nodes"
)
@click.option(
    "--show-context", is_flag=True, default=False, required=False, help="Print node context"
)
@click.option("--glyph-type", default="std", required=False, help="Tree glyph type")
def tree(view, show_context, glyph_type):
    session = create_session(hw_server_url=_hw_url, cs_server_url=_cs_url)
    tree_printer = TreePrinter(session, view)
    tree_printer.print(show_context=show_context, glyph_type=glyph_type)


def _add_commands():
    _chipscopy.add_command(tree)
    _chipscopy.add_command(program)
    _chipscopy.add_command(ls)
    _chipscopy.add_command(info)
    _chipscopy.add_command(json_devices)
    _chipscopy.add_command(report)


def main():  # pragma: no cover
    _add_commands()
    _chipscopy()


if __name__ == "__main__":  # pragma: no cover
    main()



# *****************************************************************************
# * Copyright (c) 2020 Xilinx, Inc.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Xilinx
# *****************************************************************************



# *****************************************************************************
# * Copyright (c) 2020 Xilinx, Inc.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Xilinx
# *****************************************************************************

from chipscopy.tcf.services import Service, DoneHWCommand

NAME = "svf"
"""SVF service name."""


class SVFService(Service):
    """TCF SVF service interface."""
    def getName(self) -> str:
        """
        Get this service name.

        :returns: The value of string :const:`NAME`
        """
        return NAME

    def add_target(self, name: str, done: DoneHWCommand) -> None:
        """
        Create an SVF target.

        :param name: Name of target
        :param done: Callback with result and any error.
        """
        raise NotImplementedError("Abstract method")

    def add_device(self, target_ctx: str, name: str, done: DoneHWCommand) -> str:
        """
        Adds SVF device from database using the given name to the target chain.

        :param target_ctx: Context id of the target chain
        :param name: Name of the device to add (must be a known device)
        :param done: Callback with result and any error
        :return: Name of added device
        """
        raise NotImplementedError("Abstract method")

    def remove_target(self, target_ctx: str, done: DoneHWCommand) -> None:
        """
        Deletes the target and any associated devices.

        :param target_ctx: Context id of the target chain
        :param done: Callback with the result and any error
        """
        raise NotImplementedError("Abstract method")



# *****************************************************************************
# * Copyright (c) 2020 Xilinx, Inc.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Xilinx
# *****************************************************************************

from typing import Dict, List, Union, Tuple
from chipscopy.tcf.services import mila, DoneHWCommand
from chipscopy.proxies.CorePropertyProxy import CorePropertyProxy


class MILAProxy(CorePropertyProxy, mila.MILAService):
    def __init__(self, channel):
        super().__init__(channel)
        self._listeners = {}

    #
    #  General methods
    #
    def initialize(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_command('initialize', (node_id, ), done)

    def terminate(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_command('terminate', (node_id, ), done)

    def write32(self, node_id: str, write_addr: int, data: int, domain_index: int, done: DoneHWCommand) -> None:
        return self.send_command('write32', (node_id, write_addr, data, domain_index), done)

    def read32(self, node_id: str, read_addr: int, read_word_count: int, domain_index: int, done: DoneHWCommand) -> \
            Union[List[int], int]:
        return self.send_command('read32', (node_id, read_addr, read_word_count, domain_index), done)

    #
    # Probe Methods
    #
    def define_probe(self, node_id: str, probe_defs: mila.MILAProbeDefs, done: DoneHWCommand) -> None:
        return self.send_command('defineProbe', (node_id, probe_defs,), done)

    def undefine_probe(self, node_id: str, probe_names: List[str], done: DoneHWCommand) -> None:
        return self.send_command('undefineProbe', (node_id, probe_names,), done)

    def get_probe(self, node_id: str, probe_names: List[str], done: DoneHWCommand) -> mila.MILAProbeDefs:
        return self.send_command('getProbe', (node_id, probe_names,), done)

    def get_probe_match_value(self, node_id: str, probe_names: List[str], done: DoneHWCommand)\
            -> Dict[str, str]:
        return self.send_command('getProbeMatchValue', (node_id, probe_names,), done)

    def set_probe_match_value(self, node_id: str, match_pairs: Dict[str, str], done: DoneHWCommand)\
            -> None:
        return self.send_command('setProbeMatchValue', (node_id, match_pairs,), done)

    #
    # Core Communication Methods
    #
    def arm(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_command('arm', (node_id, ), done)

    def upload(self, node_id: str, done: DoneHWCommand):
        return self.send_command('upload', (node_id,), done)

    #
    #  MUX Methods
    #
    def mux_add_node(self, node_id: str,
                     memory_domain: str,
                     addr: Union[int, None],
                     name: str,
                     long_name: str,
                     has_local_inputs: bool,
                     ch_in0: str,
                     ch_in1: str,
                     done: DoneHWCommand):
        return self.send_command('muxAddNode',
                                 (node_id, memory_domain, addr, name, long_name,
                                  has_local_inputs, ch_in0, ch_in1),
                                 done)

    def mux_build_tree(self, node_id: str,
                       clk_sel_is_1_mux: str,
                       ila_capture_data_sel_muxes: [str],
                       done: DoneHWCommand):
        return self.send_command(
            'muxBuildTree', (node_id, clk_sel_is_1_mux, ila_capture_data_sel_muxes), done)

    def mux_select_inputs(self, node_id: str, mux_node_inputs: [Tuple[str, int]],
                          done: DoneHWCommand) -> None:
        return self.send_command('muxSelectInputs', (node_id, mux_node_inputs), done)

    def mux_get_selected_inputs(self, node_id: str, done: DoneHWCommand):
        return self.send_command('muxGetSelectedInputs', (node_id,), done)

    def mux_commit(self, node_id: str, done: DoneHWCommand):
        return self.send_command('muxCommit', (node_id,), done)

    def mux_refresh(self, node_id: str, done: DoneHWCommand):
        return self.send_command('muxRefresh', (node_id,), done)

    def mux_report(self, node_id: str, skip_default: bool, done: DoneHWCommand):
        return self.send_command('muxReport', (node_id, skip_default), done)



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.10.1
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# <link rel="preconnect" href="https://fonts.gstatic.com">
# <link href="https://fonts.googleapis.com/css2?family=Fira+Code&display=swap" rel="stylesheet">
#
# ### License
#
# <p style="font-family: 'Fira Code', monospace; font-size: 1.2rem">
# Copyright 2021 Xilinx, Inc.<br><br>
# Licensed under the Apache License, Version 2.0 (the "License");<br>
# you may not use this file except in compliance with the License.<br><br>
# You may obtain a copy of the License at <a href="http://www.apache.org/licenses/LICENSE-2.0"?>http://www.apache.org/licenses/LICENSE-2.0</a><br><br>
# Unless required by applicable law or agreed to in writing, software<br>
# distributed under the License is distributed on an "AS IS" BASIS,<br>
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.<br>
# See the License for the specific language governing permissions and<br>
# limitations under the License.<br>
# </p>
#

# %% [markdown]
# # IBERT link and eye scan example
#
#
# <img src="../../img/api_overview.png" width="500" align="left">

# %% [markdown]
# ## Description
# This example shows how to interact with the IBERT (Integrated Bit Error Ratio Tester) debug core service via ChipScoPy APIs.
# - Program the ChipScoPy CED design onto the XCVC1902 device on a VCK190 board
# - Verify that the expected IBERT quads are instantiated by the design
# - Create links and change link settings to get link lock
# - Run and plot eye scans for the links
#
#
# ## Requirements
# - Local or remote Xilinx Versal board, such as a VCK190
# - Xilinx hw_server 2021.2 installed and running
# - Xilinx cs_server 2021.2 installed and running
# - Python 3.8 or greater installed
# - ChipScoPy 2021.2 installed
# - Jupyter notebook support installed - Please do so, using the command `pip install chipscopy[jupyter]`
# - Plotting support installed - Please do so, using the command `pip install chipscopy[core-addons]`

# %% [markdown]
# ## 1 - Initialization: Imports and File Paths
#
# After this step,
# - Required functions and classes are imported
# - Paths to server(s) and files are set correctly

# %% jupyter={"outputs_hidden": false} pycharm={"name": "#%%\n"}
import os
from more_itertools import one

from chipscopy import create_session, report_versions, report_hierarchy, get_design_files
from chipscopy.api.ibert.aliases import (
    EYE_SCAN_HORZ_RANGE,
    EYE_SCAN_VERT_RANGE,
    EYE_SCAN_VERT_STEP,
    EYE_SCAN_HORZ_STEP,
    EYE_SCAN_TARGET_BER,
    PATTERN,
    RX_LOOPBACK,
)
from chipscopy.api.ibert import create_eye_scans, create_links


# %%
# Make sure to start the hw_server and cs_server prior to running.
# Specify locations of the running hw_server and cs_server below.
# The default is localhost - but can be other locations on the network.
CS_URL = os.getenv("CS_SERVER_URL", "TCP:localhost:3042")
HW_URL = os.getenv("HW_SERVER_URL", "TCP:localhost:3121")

# The get_design_files() function tries to find the PDI and LTX files. In non-standard
# configurations, you can put the path for PROGRAMMING_FILE and PROBES_FILE below.
design_files = get_design_files("vck190/production/chipscopy_ced")

PDI_FILE = design_files.programming_file

print(f"PROGRAMMING_FILE: {PDI_FILE}")

# %% [markdown]
# ## 2 - Create a session and connect to the hw_server and cs_server
#
# The session is a container that keeps track of devices and debug cores.
# After this step,
# - Session is initialized and connected to server(s)
# - Versions are detected and reported to stdout

# %% jupyter={"outputs_hidden": false} pycharm={"name": "#%%\n"}
session = create_session(cs_server_url=CS_URL, hw_server_url=HW_URL)
report_versions(session)

# %% [markdown]
# ## 3 - Program the device with the example design
#
# After this step,
# - Device is programmed with the example programming file

# %% jupyter={"outputs_hidden": false} pycharm={"name": "#%%\n"}
# Typical case - one device on the board - get it.
device = session.devices.filter_by(family="versal").get()
device.program(PDI_FILE)

# %% [markdown]
# ## 4 - Discover and setup the IBERT core
#
# Debug core discovery initializes the chipscope server debug cores.
#
# After this step,
#
# - The cs_server is initialized and ready for use
# - The first ibert found is used

# %% jupyter={"outputs_hidden": false} pycharm={"name": "#%%\n"}
device.discover_and_setup_cores(ibert_scan=True)
print("--> Debug core discovery done")

if len(device.ibert_cores) == 0:
    print("No IBERT core found! Exiting...")
    exit()

# Use the first available IBERT core from the device
ibert = device.ibert_cores.at(index=0)

if len(ibert.gt_groups) == 0:
    print("No GT Groups available for use! Exiting...")
    exit()

print(f"GT Groups available - {[gt_group_obj.name for gt_group_obj in ibert.gt_groups]}")

# %% [markdown]
# ## 5 - Print the hierarchy of the IBERT core
#
# We also ensure that all the quads instantiated by the ChipScoPy CED design are found by the APIs

# %% jupyter={"outputs_hidden": false} pycharm={"name": "#%%\n"}
report_hierarchy(ibert)

assert len(ibert.gt_groups) == 4

q205 = one(ibert.gt_groups.filter_by(name="Quad_205"))
q204 = one(ibert.gt_groups.filter_by(name="Quad_204"))
q201 = one(ibert.gt_groups.filter_by(name="Quad_201"))
q105 = one(ibert.gt_groups.filter_by(name="Quad_105"))

# %% [markdown]
# ## 6 - Create links between following TXs and RXs
# - Quad 205 CH0 TX to Quad 205 CH0 RX
# - Quad 204 CH1 TX to Quad 204 CH1 RX
# - Quad 201 CH2 TX to Quad 201 CH2 RX
# - Quad 105 CH3 TX to Quad 105 CH3 RX

# %% pycharm={"name": "#%%\n"}
links = create_links(
    txs=[q205.gts[0].tx, q204.gts[1].tx, q201.gts[2].tx, q105.gts[3].tx],
    rxs=[q205.gts[0].rx, q204.gts[1].rx, q201.gts[2].rx, q105.gts[3].rx],
)

print("--> Done creating links")

# %% [markdown] raw_mimetype="text/markdown"
# ## 7 - Print the valid values for pattern and loopback, set the pattern for the TXs and RXs to "PRBS 31" and set loopback to "Near-End PMA"
#
# In order to lock the internal pattern checker, TX and RX patterns need to match. We also need to have some kind of loopback, internal/external.
#
# We are assuming that no external cable loopback is present and hence making use of internal loopback.

# %% pycharm={"name": "#%%\n"} raw_mimetype="text/x-python"
for link in links:
    print(f"\n----- {link.name} -----")
    _, tx_pattern_report = link.tx.property.report(link.tx.property_for_alias[PATTERN]).popitem()
    _, rx_pattern_report = link.rx.property.report(link.rx.property_for_alias[PATTERN]).popitem()
    _, rx_loopback_report = link.tx.property.report(
        link.rx.property_for_alias[RX_LOOPBACK]
    ).popitem()

    print(f"--> Valid values for TX pattern - {tx_pattern_report['Valid values']}")
    print(f"--> Valid values for RX pattern - {rx_pattern_report['Valid values']}")
    print(f"--> Valid values for RX loopback - {rx_loopback_report['Valid values']}")

    props = {link.tx.property_for_alias[PATTERN]: "PRBS 31"}
    link.tx.property.set(**props)
    link.tx.property.commit(list(props.keys()))

    props = {
        link.rx.property_for_alias[PATTERN]: "PRBS 31",
        link.rx.property_for_alias[RX_LOOPBACK]: "Near-End PMA",
    }
    link.rx.property.set(**props)
    link.rx.property.commit(list(props.keys()))
    print(f"\n--> Set both patterns to 'PRBS 31' & loopback to 'Near-End PMA' for {link}")

    assert link.rx.pll.locked and link.tx.pll.locked
    print(f"--> RX and TX PLLs are locked for {link}. Checking for link lock...")
    assert link.status != "No link"
    print(f"--> {link} is linked as expected")

# %% [markdown]
# ## 8 - Create eye scan objects for all the links, set the scan params and start the scan
#
# The eye scans will be run in parallel

# %% pycharm={"name": "#%%\n"}
eye_scans = create_eye_scans(target_objs=[link for link in links])
for eye_scan in eye_scans:
    eye_scan.params[EYE_SCAN_HORZ_STEP].value = 2
    eye_scan.params[EYE_SCAN_VERT_STEP].value = 2
    eye_scan.params[EYE_SCAN_HORZ_RANGE].value = "-0.500 UI to 0.500 UI"
    eye_scan.params[EYE_SCAN_VERT_RANGE].value = "100%"
    eye_scan.params[EYE_SCAN_TARGET_BER].value = 1e-5

    eye_scan.start()
    print(f"Started eye scan {eye_scan}")


# %% [markdown]
# ## 9 - Wait for all the eye scans to get done

# %% pycharm={"name": "#%%\n"}
eye_scans[0].wait_till_done()
eye_scans[1].wait_till_done()
eye_scans[2].wait_till_done()
eye_scans[3].wait_till_done()

# %% [markdown]
# ## 10 - View Eye Scan Plot.
#
# This requires Plotly to be installed. See how to install it [here](https://pages.gitenterprise.xilinx.com/chipscope/chipscopy/2020.2/ibert/scan.html#scan-plots)
#
# NOTE - The plot may not display if this notebook is run in Jupyter Lab. For details, see [link](https://plotly.com/python/getting-started/#jupyterlab-support-python-35)


# %% pycharm={"name": "#%%\n"}
eye_scans[0].plot.show()
eye_scans[1].plot.show()
eye_scans[2].plot.show()
eye_scans[3].plot.show()



# *****************************************************************************
# * Copyright (c) 2020 Xilinx, Inc.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License v1.0
# * which accompanies this distribution, and is available at
# * http://www.eclipse.org/legal/epl-v10.html
# *
# * Contributors:
# *     Xilinx
# *****************************************************************************

from chipscopy.tcf.services.arguments import *

# from ... import errors, channel
from chipscopy.tcf.channel.Command import Command
from chipscopy.tcf.services import xicom


class XicomEverestProxy(xicom.XicomEverestService):
    def __init__(self, channel):
        super(XicomEverestProxy, self).__init__(channel)
        self.listeners = {}

    def config(self, ctx, data, options={}, done=None):
        args = {'ctx': ctx, "data": data}
        args.update(options)
        return self.send_xicom_command("config", args, done)

    def get_status(self, ctx, need_definitions, done):
        args = {'ctx': ctx, "need_definitions": need_definitions}
        return self.send_xicom_command("status", args, done)

    def get_static(self, ctx, need_definitions, done):
        args = {'ctx': ctx, "need_definitions": need_definitions}
        return self.send_xicom_command("get_static", args, done)

    def reset(self, ctx, done):
        args = {'ctx': ctx}
        return self.send_xicom_command("reset", args, done)

    def test(self, props, done):
        return self.send_xicom_command("test", props, done)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Dict, Any, List
from chipscopy.tcf import channel
from chipscopy.tcf.services import Service, DoneHWCommand
from chipscopy.tcf.services.arguments import from_xargs
from chipscopy.utils.logger import log
from chipscopy.utils import words_from_bytes


NAME = "DebugCore"
"""DebugCore service name."""


class DebugNodeListener(object):
    """Debug node event listener is notified when debug node hierarchy changes,
    and when node is modified by DebugCore service commands.
    """

    def node_added(self, node_ctx: str, props: Dict[str, Any]):
        """
        Called when a new Debug Node was added
        :param node_ctx: Debug node context id
        :param props: Debug node properties added
        """
        pass

    def node_changed(self, node_ctx: str, props: Dict[str, Any]):
        """
        Called when a new Debug Node was changed
        :param node_ctx: Debug node context id
        :param props: Debug node properties changed
        """
        pass

    def node_removed(self, node_ctx: str):
        """
        Called when a new Debug Node was removed
        :param node_ctx: Debug node context id
        """
        pass


class DebugCoreProxy(Service):
    def __init__(self, channel):
        super().__init__(channel)
        self.listeners = {}

    """TCF DebugCore service interface."""

    def getName(self):
        """Get this service name.

        :returns: The value of string :const:`NAME`
        """
        return NAME

    def detect_hubs(self, ctx: str, hub_addresses: List[int], done: DoneHWCommand):
        """Detects cores given a parent context and expected core offsets

        :param ctx: Parent context handle
        :param hub_addresses: List of offsets for each Debug Hub
        :param done: Callback with result and any error
        :returns: Token of command request
        """
        if not hub_addresses:
            hub_addresses = []
        return self.send_xicom_command("detectHubs", (ctx, hub_addresses), done)

    def setup_cores(self, ctx: str, done: DoneHWCommand):
        """
        Sets up hardened debug nodes from a context
        :param ctx: Context underwhich to setup debug nodes
        :param done: Callback with result and any error
        :returns: Token of command request
        """
        return self.send_xicom_command("setupCores", (ctx,), done)

    def remove_cores(self, ctx: str, done: DoneHWCommand):
        """
        Removes child cores from a Debug node context
        :param ctx: Context underwhich to remove debug nodes
        :param done: Callback with result and any error
        :returns: Token of command request
        """
        return self.send_xicom_command("removeCores", (ctx,), done)

    def get_children(self, ctx: str, done: DoneHWCommand):
        """Retrieves list of children given a parent context.
        If no parent context given it returns the available list of parent core contexts.

        :param ctx: Parent context handle (defaults None)
        :param done: Callback with result and any error
        :returns: Token of command request
        :results: children - list of children
        """
        if not ctx:
            ctx = ""
        return self.send_xicom_command("getChildren", (ctx,), done)

    def get_context(self, ctx: str, done: DoneHWCommand):
        """Retrieves list of context given a debug core context.

        :param ctx: Debug core context
        :param done: Callback with result and any error
        :returns: Token of command request
        :results: |Dict| of available context properties
        """
        return self.send_xicom_command("getContext", (ctx,), done)

    def run_sequence(self, ctx: str, seq, done: DoneHWCommand):
        """
        Runs a sequence of memory transactions.

        :param ctx: Debug core context
        :param seq: List of transactions to run in the given sequence.
            Each transaction is a dict of transaction properties.
        :param done: Callback with result and any error
        :returns: Token of command request
        :results: seq - list of read transactions with resulting values in the order ran.
        """
        if log.is_domain_enabled("debugcore", "DEBUG"):
            log.debugcore.debug(f"Running sequence {len(seq)}:")
            for op in seq:
                log.debugcore.debug(f"\t{op}")
                if "data" in op:
                    words = words_from_bytes(op["data"])
                    hexstr = " ".join(f"{word:08X}" for word in words)
                    log.debugcore.debug(f"\t\t{hexstr}")
        return self.send_xicom_command("sequence", (ctx, seq), done)

    def lock(self, ctx: str, done: DoneHWCommand):
        """Locks a targeted debug core so that only this channel can access.
        Debug cores automatically unlock during a channel disconnect.

        :param ctx: Debug core context to lock
        :param done: Callback with result and any error
        :returns: Token of command request
        :results: success - true if locked
        """
        raise NotImplementedError("Abstract method")

    def unlock(self, ctx: str, done: DoneHWCommand):
        """Unlocks a targeted debug core. See `lock`

        :param ctx: Debug core context to unlock
        :param done: Callback with result and any error
        :returns: Token of command request
        :results: success - true if unlocked.  Also considered successful if already unlocked.
        """
        raise NotImplementedError("Abstract method")

    def perf_test(self, ctx: str, props: Dict[str, Any] = {}, done: DoneHWCommand = None):
        return self.send_xicom_command("perfTest", (ctx, props), done)

    def add_listener(self, listener: DebugNodeListener):
        """Add DebugCore service event listener.
        :param listener: Event listener implementation.
        """
        l = ChannelEventListener(self, listener)
        self.channel.addEventListener(self, l)
        self.listeners[listener] = l

    def remove_listener(self, listener: DebugNodeListener):
        """Remove DebugCore service event listener.
        :param listener: Event listener implementation.
        """
        l = self.listeners.get(listener)
        if l:
            del self.listeners[listener]
            self.channel.removeEventListener(self, l)


class ChannelEventListener(channel.EventListener):
    def __init__(self, service, listener):
        self.service = service
        self.listener = listener

    def event(self, name, data):
        try:
            args = channel.fromJSONSequence(data)
            if name == "nodeAdded":
                args = from_xargs(args)
                assert len(args) >= 2
                self.listener.node_added(args[0], args[1])
            elif name == "nodeChanged":
                args = from_xargs(args)
                assert len(args) >= 2
                self.listener.node_changed(args[0], args[1])
            elif name == "nodeRemoved":
                args = from_xargs(args)
                assert len(args) == 1
                self.listener.node_removed(args[0])
            else:
                raise IOError("Memory service: unknown event: " + name)
        except Exception as x:
            import sys

            x.tb = sys.exc_info()[2]
            self.service.channel.terminate(x)


DebugCoreService = DebugCoreProxy



# *****************************************************************************
# * Copyright (c) 2011, 2013-2014, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

import sys
import time
import threading
import atexit
import tcf  # @UnresolvedImport

from tcf import protocol, channel, errors  # @UnresolvedImport
from tcf.util import sync  # @UnresolvedImport

__TRACE = False


class TraceListener(channel.TraceListener):

    def onMessageReceived(self, msgType, token, service, name, data):
        print("<<< ", msgType, token, service, name, data)

    def onMessageSent(self, msgType, token, service, name, data):
        print(">>>", msgType, token, service, name, data)

    def onChannelClosed(self, error):
        sys.stderr.write("*** closed *** " + str(error) + "\n")

_suspended = []
_memory = []
_services = []


def test():
    global _services
    protocol.startEventQueue()
    atexit.register(protocol.getEventQueue().shutdown)
    # testTimer()
    try:
        c = tcf.connect("TCP:127.0.0.1:1534")
    except Exception as e:
        protocol.log(e)
        sys.exit()
    assert c.state == channel.STATE_OPEN
    if __TRACE:
        protocol.invokeAndWait(c.addTraceListener, TraceListener())
    _services = sorted(protocol.invokeAndWait(c.getRemoteServices))
    print("services=" + str(_services))

    if "RunControl" in _services:
        # RunControl must be first
        _services.remove("RunControl")
        _services.insert(0, "RunControl")
    for service in _services:
        testFct = globals().get("test" + service)
        if testFct:
            print("Testing service '%s'..." % service)
            try:
                testFct(c)
                print("Completed test of service '%s'." % service)
            except Exception as e:
                protocol.log("Exception testing %s" % service, e)
        else:
            print("No test for service '%s' found." % service)
    try:
        testSyncCommands(c)
        testTasks(c)
        testEvents(c)
        testDataCache(c)
    except Exception as e:
        protocol.log(e)

    if c.state == channel.STATE_OPEN:
        time.sleep(5)
        protocol.invokeLater(c.close)
    time.sleep(2)


def testTimer():
    cond = threading.Condition()

    def countdown(left):
        if left == 0:
            print("Ignition sequence started!")
            with cond:
                cond.notify()
            return
        print("%d seconds to go" % left)
        sys.stdout.flush()
        protocol.invokeLaterWithDelay(1000, countdown, left - 1)
    with cond:
        protocol.invokeLaterWithDelay(0, countdown, 10)
        cond.wait(15)


def testRunControl(c):
    lock = threading.Condition()
    from tcf.services import runcontrol  # @UnresolvedImport

    def getContexts():
        rctrl = c.getRemoteService(runcontrol.NAME)
        pending = []

        class DoneGetContext(runcontrol.DoneGetContext):
            def doneGetContext(self, token, error, context):
                pending.remove(token)
                if error:
                    protocol.log("Error from RunControl.getContext", error)
                else:
                    print(context)

                class DoneGetState(runcontrol.DoneGetState):

                    def doneGetState(self, token, error, suspended, pc, reason,
                                     params):
                        pending.remove(token)
                        if error:
                            protocol.log(
                                "Error from RunControl.getState", error)
                        else:
                            print("suspended: " + str(suspended))
                            print("pc:        " + str(pc))
                            print("reason:    " + str(reason))
                            print("params:    " + str(params))
                        if suspended:
                            _suspended.append(context.getID())
                        if len(pending) == 0:
                            with lock:
                                lock.notify()
                if context and context.hasState():
                    pending.append(context.getState(DoneGetState()))
                if len(pending) == 0:
                    with lock:
                        lock.notify()

        class DoneGetChildren(runcontrol.DoneGetChildren):
            def doneGetChildren(self, token, error, context_ids):
                pending.remove(token)
                if error:
                    protocol.log("Error from RunControl.GetChildren", error)
                else:
                    for c in context_ids:
                        pending.append(rctrl.getContext(c, DoneGetContext()))
                        pending.append(rctrl.getChildren(c, self))
                if len(pending) == 0:
                    with lock:
                        lock.notify()
        pending.append(rctrl.getChildren(None, DoneGetChildren()))
    with lock:
        protocol.invokeLater(getContexts)
        lock.wait(5)

    def listenerTest():
        rc = c.getRemoteService(runcontrol.NAME)

        class RCListener(runcontrol.RunControlListener):
            def contextSuspended(self, *args):
                print("context suspended: " + str(args))
                rc.removeListener(self)

            def contextResumed(self, *args):
                print("context resumed: " + str(args))

            def containerSuspended(self, *args):
                print("container suspended: " + str(args))
                rc.removeListener(self)

            def containerResumed(self, *args):
                print("container resumed: " + str(args))
        rc.addListener(RCListener())

        class DoneGetContext(runcontrol.DoneGetContext):
            def doneGetContext(self, token, error, context):
                if error:
                    protocol.log("Error from RunControl.getContext", error)
                    with lock:
                        lock.notify()
                    return

                class DoneResume(runcontrol.DoneCommand):
                    def doneCommand(self, token, error):
                        if error:
                            protocol.log("Error from RunControl.resume", error)
                        else:
                            context.suspend(runcontrol.DoneCommand())
                        with lock:
                            lock.notify()
                context.resume(runcontrol.RM_RESUME, 1, None, DoneResume())
        rc.getContext(_suspended[0], DoneGetContext())

    if _suspended:
        with lock:
            protocol.invokeLater(listenerTest)
            lock.wait(5)
        _suspended.pop(0)


def testBreakpoints(c):
    from tcf.services import breakpoints  # @UnresolvedImport

    def testBPQuery():
        bps = c.getRemoteService(breakpoints.NAME)

        def doneGetIDs(token, error, ids):
            if error:
                protocol.log("Error from Breakpoints.getIDs", error)
                return
            print("Breakpoints : " + str(ids))

            def doneGetProperties(token, error, props):
                if error:
                    protocol.log("Error from Breakpoints.getProperties", error)
                    return
                print("Breakpoint Properties: " + str(props))

            def doneGetStatus(token, error, props):
                if error:
                    protocol.log("Error from Breakpoints.getStatus", error)
                    return
                print("Breakpoint Status: " + str(props))
            for bpid in ids:
                bps.getProperties(bpid, doneGetProperties)
                bps.getStatus(bpid, doneGetStatus)
        bps.getIDs(doneGetIDs)
    protocol.invokeLater(testBPQuery)

    def testBPSet():
        bpsvc = c.getRemoteService(breakpoints.NAME)

        class BPListener(breakpoints.BreakpointsListener):
            def breakpointStatusChanged(self, bpid, status):
                print("breakpointStatusChanged " + str(bpid) + " " +
                      str(status))

            def contextAdded(self, bps):
                print("breakpointAdded " + str(bps))
                bpsvc.removeListener(self)

            def contextChanged(self, bps):
                print("breakpointChanged " + str(bps))

            def contextRemoved(self, ids):
                print("breakpointRemoved " + str(ids))
        bpsvc.addListener(BPListener())

        def doneSet(token, error):
            if error:
                protocol.log("Error from Breakpoints.set", error)
                return
        bp = {
            breakpoints.PROP_ID: "python:1",
            breakpoints.PROP_ENABLED: True,
            breakpoints.PROP_LOCATION: "sysClkRateGet"
        }
        bpsvc.set([bp], doneSet)
    protocol.invokeLater(testBPSet)


def testStackTrace(c):
    from tcf.services import stacktrace  # @UnresolvedImport

    def stackTest(ctx_id):
        stack = c.getRemoteService(stacktrace.NAME)

        class DoneGetChildren(stacktrace.DoneGetChildren):
            def doneGetChildren(self, token, error, ctx_ids):
                if error:
                    protocol.log("Error from StackTrace.getChildren", error)
                    return

                class DoneGetContext(stacktrace.DoneGetContext):
                    def doneGetContext(self, token, error, ctxs):
                        if error:
                            protocol.log(
                                "Error from StackTrace.getContext", error)
                            return
                        if ctxs:
                            for ctx in ctxs:
                                print(ctx)
                stack.getContext(ctx_ids, DoneGetContext())
        stack.getChildren(ctx_id, DoneGetChildren())
    for ctx_id in _suspended:
        protocol.invokeLater(stackTest, ctx_id)


def testDisassembly(c):
    if not _suspended:
        return
    ctl = sync.CommandControl(c)
    try:
        dis = ctl.Disassembly
    except AttributeError:
        # no Disassembly service
        return
    for ctx_id in _suspended:
        frames = ctl.StackTrace.getChildren(ctx_id).get()
        if frames:
            frameData = ctl.StackTrace.getContext(frames).get()
            if frameData:
                addr = frameData[0].get("IP")
                if addr:
                    print("Disassemble context %s from 0x%x" % (ctx_id, addr))
                    lines = dis.disassemble(ctx_id, addr, 256, None).get()
                    if lines:
                        for line in lines:
                            print(line)


def testSymbols(c):
    from tcf.services import symbols  # @UnresolvedImport

    def symTest(ctx_id):
        syms = c.getRemoteService(symbols.NAME)

        class DoneList(symbols.DoneList):
            def doneList(self, token, error, ctx_ids):
                if error:
                    protocol.log("Error from Symbols.list", error)
                    return

                class DoneGetContext(symbols.DoneGetContext):
                    def doneGetContext(self, token, error, ctx):
                        if error:
                            protocol.log(
                                "Error from Symbols.getContext", error)
                            return
                        print(ctx)
                if ctx_ids:
                    for ctx_id in ctx_ids:
                        syms.getContext(ctx_id, DoneGetContext())
        syms.list(ctx_id, DoneList())
    for ctx_id in _suspended:
        protocol.invokeLater(symTest, ctx_id)


def testRegisters(c):
    if not _suspended:
        return
    from tcf.services import registers  # @UnresolvedImport
    lock = threading.Condition()

    def regTest(ctx_id):
        regs = c.getRemoteService(registers.NAME)
        pending = []

        def onDone():
            with lock:
                lock.notify()

        class DoneGetChildren(registers.DoneGetChildren):
            def doneGetChildren(self, token, error, ctx_ids):
                pending.remove(token)
                if error:
                    protocol.log("Error from Registers.getChildren", error)
                if not pending:
                    onDone()

                class DoneGetContext(registers.DoneGetContext):
                    def doneGetContext(self, token, error, ctx):
                        pending.remove(token)
                        if error:
                            protocol.log(
                                "Error from Registers.getContext", error)
                        else:
                            print(ctx)
                            if ctx.isReadable() and not ctx.isReadOnce() \
                                    and ctx.getSize() >= 2:
                                locs = []
                                locs.append(
                                    registers.Location(ctx.getID(), 0, 1))
                                locs.append(
                                    registers.Location(ctx.getID(), 1, 1))

                                class DoneGetM(registers.DoneGet):
                                    def doneGet(self, token, error, value):
                                        pending.remove(token)
                                        if error:
                                            protocol.log(
                                                "Error from Registers.getm",
                                                error)
                                        else:
                                            print("getm " + str(ctx.getID()) +
                                                  " " +
                                                  str(list(map(int, value))))
                                        if not pending:
                                            onDone()
                                pending.append(regs.getm(locs, DoneGetM()))
                            if ctx.isWriteable() and not ctx.isWriteOnce() \
                                    and ctx.getSize() >= 2:
                                locs = []
                                locs.append(
                                    registers.Location(ctx.getID(), 0, 1))
                                locs.append(
                                    registers.Location(ctx.getID(), 1, 1))

                                class DoneSetM(registers.DoneSet):
                                    def doneGet(self, token, error):
                                        pending.remove(token)
                                        if error:
                                            protocol.log(
                                                "Error from Registers.setm",
                                                error)
                                        if not pending:
                                            onDone()
                                pending.append(
                                    regs.setm(locs, (255, 255), DoneSetM()))
                        if not pending:
                            onDone()
                if ctx_ids:
                    for ctx_id in ctx_ids:
                        pending.append(
                            regs.getContext(ctx_id, DoneGetContext()))
        pending.append(regs.getChildren(ctx_id, DoneGetChildren()))
    with lock:
        for ctx_id in _suspended:
            protocol.invokeLater(regTest, ctx_id)
        lock.wait(5)


def testExpressions(c):
    if not _suspended:
        return
    from tcf.services import expressions  # @UnresolvedImport
    ctl = sync.CommandControl(c)
    exprs = ctl.Expressions
    e = exprs.create(_suspended[0], None, "1+2*(3-4/2)").getE()
    eid = e.get(expressions.PROP_ID)
    val, cls = exprs.evaluate(eid).getE()
    print(e.get(expressions.PROP_EXPRESSION) + " = " + str(val))
    exprs.dispose(eid)


def testLineNumbers(c):
    if not _suspended:
        return
    from tcf.services import stacktrace  # @UnresolvedImport
    ctl = sync.CommandControl(c)
    stack = ctl.StackTrace
    lineNumbers = ctl.LineNumbers
    for ctx_id in _suspended:
        bt = stack.getChildren(ctx_id).get()
        if bt:
            bt = stack.getContext(bt).get()
            for frame in bt:
                addr = frame.get(stacktrace.PROP_INSTRUCTION_ADDRESS) or 0
                area = lineNumbers.mapToSource(ctx_id, addr, addr + 1).get()
                print("Frame %d - CodeArea: %s" %
                      (frame.get(stacktrace.PROP_LEVEL), area))


def testSyncCommands(c):
    # simplified command execution
    ctl = sync.CommandControl(c)
    try:
        diag = ctl.Diagnostics
    except AttributeError:
        # no Diagnostics service
        return
    s = "Hello TCF World"
    r = diag.echo(s).getE()
    assert s == r
    pi = 3.141592654
    r = diag.echoFP(pi).getE()
    assert pi == r
    e = errors.ErrorReport("Test", errors.TCF_ERROR_OTHER)
    r = diag.echoERR(e.getAttributes()).getE()
    assert e.getAttributes() == r
    print("Diagnostic tests: " + str(diag.getTestList().getE()))

    for ctx_id in _suspended:
        print("Symbols: " + str(ctl.Symbols.list(ctx_id)))
    for ctx_id in _suspended:
        frame_ids = ctl.StackTrace.getChildren(ctx_id).get()
        if frame_ids:
            error, args = ctl.StackTrace.getContext(frame_ids)
            if not error:
                print("Stack contexts: " + str(args))
    try:
        ctl.Breakpoints
    except AttributeError:
        # no Breakpoints service
        return

    def gotBreakpoints(error, bps):
        print("Got breakpoint list: " + str(bps))
    ctl.Breakpoints.getIDs(onDone=gotBreakpoints)
    try:
        print(ctl.Processes.getChildren(None, False))
    except:
        pass  # no Processes service


def testTasks(c):
    if not _suspended:
        return
    from tcf.services import expressions  # @UnresolvedImport
    from tcf.util import task  # @UnresolvedImport

    def compute(expr, done=None):
        es = c.getRemoteService(expressions.NAME)
        if not es:
            done(Exception("No Expressions service"), None)
            return

        def doneCreate(token, error, ctx):
            if error:
                done(error, None)
                return

            def doneEval(token, error, val):
                done(error, val)
            es.evaluate(ctx.getID(), doneEval)
        es.create(_suspended[0], None, expr, doneCreate)
    t = task.Task(compute, "1+2*(3-4/2)", channel=c)
    val = t.get()
    print("Task result: " + str(val))


def testEvents(c):
    from tcf.util import event  # @UnresolvedImport
    recorder = event.EventRecorder(c)
    recorder.record("RunControl")
    ctl = sync.CommandControl(c)
    try:
        rc = ctl.RunControl
    except AttributeError:
        # no RunControl service
        return
    ctxs = rc.getChildren(None).get()
    if not ctxs:
        return
    ctx = ctxs[0]
    rc.resume(ctx, 0, 1, None).wait()
    print(recorder)
    rc.suspend(ctx).wait()
    print(recorder)
    recorder.stop()


def testDataCache(c):
    from tcf.util import cache  # @UnresolvedImport
    from tcf.services import runcontrol  # @UnresolvedImport
    if runcontrol.NAME not in _services:
        return

    class ContextsCache(cache.DataCache):
        def startDataRetrieval(self):
            rc = self._channel.getRemoteService(runcontrol.NAME)
            if not rc:
                self.set(None, Exception("No RunControl service"), None)
                return
            cache = self
            pending = []
            contexts = []

            class DoneGetChildren(runcontrol.DoneGetChildren):
                def doneGetChildren(self, token, error, context_ids):
                    pending.remove(token)
                    if error:
                        protocol.log(
                            "Error from RunControl.GetChildren", error)
                    else:
                        for c in context_ids:
                            contexts.append(c)
                            pending.append(rc.getChildren(c, self))
                    if len(pending) == 0:
                        cache.set(None, None, contexts)
            pending.append(rc.getChildren(None, DoneGetChildren()))
    contextsCache = ContextsCache(c)

    def done():
        print("ContextsCache is valid: " + str(contextsCache.getData()))
    protocol.invokeLater(contextsCache.validate, done)


def testProcesses(c):
    from tcf.services import processes, processes_v1  # @UnresolvedImport
    lock = threading.Condition()

    def processTest():
        proc = c.getRemoteService(processes_v1.NAME) or \
            c.getRemoteService(processes.NAME)
        if not proc:
            with lock:
                lock.notify()
            return

        class DoneGetChildren(processes.DoneGetChildren):
            def doneGetChildren(self, token, error, context_ids):
                if error:
                    protocol.log("Error from Processes.GetChildren", error)
                else:
                    print("Processes: " + str(context_ids))
                with lock:
                    lock.notify()
        proc.getChildren(None, False, DoneGetChildren())
    with lock:
        protocol.invokeLater(processTest)
        lock.wait(5)


def testFileSystem(c):
    cmd = sync.CommandControl(c)
    try:
        fs = cmd.FileSystem
    except AttributeError:
        # no FileSystem service
        return
    roots = fs.roots().get()
    print("FileSystem roots: " + str(roots))
    user = fs.user().get()
    print("User info: " + str(user))


def testMemory(c):
    lock = threading.Condition()
    from tcf.services import memory  # @UnresolvedImport

    def getContexts():
        mem = c.getRemoteService(memory.NAME)
        pending = []

        class DoneGetContext(memory.DoneGetContext):
            def doneGetContext(self, token, error, context):
                pending.remove(token)
                if error:
                    protocol.log("Error from Memory.getContext", error)
                else:
                    print(context)
                if len(pending) == 0:
                    with lock:
                        lock.notify()

        class DoneGetChildren(memory.DoneGetChildren):
            def doneGetChildren(self, token, error, context_ids):
                pending.remove(token)
                if error:
                    protocol.log("Error from Memory.GetChildren", error)
                else:
                    for c in context_ids:
                        _memory.append(c)
                        pending.append(mem.getContext(c, DoneGetContext()))
                        pending.append(mem.getChildren(c, self))
                if len(pending) == 0:
                    with lock:
                        lock.notify()
        pending.append(mem.getChildren(None, DoneGetChildren()))
    with lock:
        protocol.invokeLater(getContexts)
        lock.wait(5)


def testMemoryMap(c):
    if not _memory:
        return
    cmd = sync.CommandControl(c)
    try:
        mm = cmd.MemoryMap
    except AttributeError:
        # no MemoryMap service
        return
    map_id = _memory[0]
    lock = threading.Condition()
    from tcf.services import memorymap  # @UnresolvedImport

    def getMap():
        mm = c.getRemoteService(memorymap.NAME)

        class DoneGet(memorymap.DoneGet):
            def doneGet(self, token, error, mmap):
                if error:
                    protocol.log("Error from MemoryMap.get", error)
                else:
                    print(mmap)
                with lock:
                    lock.notify()
        mm.get(map_id, DoneGet())
    with lock:
        protocol.invokeLater(getMap)
        lock.wait(1)

    def setMap():
        mm = c.getRemoteService(memorymap.NAME)

        class DoneSet(memorymap.DoneSet):
            def doneSet(self, token, error):
                if error:
                    protocol.log("Error from MemoryMap.set", error)
                with lock:
                    lock.notify()
        mm.set(map_id, {memorymap.PROP_FILE_NAME: "/tmp/system.elf"},
               DoneSet())
    with lock:
        protocol.invokeLater(setMap)
        lock.wait(1)
    mmap = mm.get(map_id).get()
    print("Memory map: " + str(mmap))


def testPathMap(c):
    cmd = sync.CommandControl(c)
    try:
        pm = cmd.PathMap
    except AttributeError:
        # no PathMap service
        return
    lock = threading.Condition()
    from tcf.services import pathmap  # @UnresolvedImport

    def getMap():
        pm = c.getRemoteService(pathmap.NAME)

        class DoneGet(pathmap.DoneGet):
            def doneGet(self, token, error, mmap):
                if error:
                    protocol.log("Error from PathMap.get", error)
                else:
                    print(mmap)
                with lock:
                    lock.notify()
        pm.get(DoneGet())
    with lock:
        protocol.invokeLater(getMap)
        lock.wait(1)

    def setMap():
        pm = c.getRemoteService(pathmap.NAME)

        class DoneSet(pathmap.DoneSet):
            def doneSet(self, token, error):
                if error:
                    protocol.log("Error from PathMap.set", error)
                with lock:
                    lock.notify()
        pm.set({pathmap.PROP_SOURCE: "/tmp",
                pathmap.PROP_DESTINATION: "/home"}, DoneSet())
    with lock:
        protocol.invokeLater(setMap)
        lock.wait(1)
    mmap = pm.get().get()
    print("Path map: " + str(mmap))


def testSysMonitor(c):
    cmd = sync.CommandControl(c)
    try:
        sm = cmd.SysMonitor
    except AttributeError:
        # no SysMonotor service
        return
    lock = threading.Condition()
    from tcf.services import sysmonitor  # @UnresolvedImport
    processes = []

    def getProcesses():
        sm = c.getRemoteService(sysmonitor.NAME)
        pending = []

        class DoneGetChildren(sysmonitor.DoneGetChildren):
            def doneGetChildren(self, token, error, context_ids):
                pending.remove(token)
                if error:
                    protocol.log("Error from SysMonitor.getChildren", error)
                else:
                    class DoneGetContext(sysmonitor.DoneGetContext):
                        def doneGetContext(self, token, error, context):
                            pending.remove(token)
                            if error:
                                protocol.log(
                                    "Error from SysMonitor.getContext", error)
                            else:
                                processes.append(context)
                            if not pending:
                                with lock:
                                    lock.notify()
                    for ctx_id in context_ids:
                        pending.append(
                            sm.getContext(ctx_id, DoneGetContext()))
                if not pending:
                    with lock:
                        lock.notify()
        pending.append(sm.getChildren(None, DoneGetChildren()))
    with lock:
        protocol.invokeLater(getProcesses)
        lock.wait(5)
    print("%d processes found:" % len(processes))
    for p in processes:
        print(p)
        cmdl = sm.getCommandLine(p.getID()).get()
        if cmdl:
            print("Command line: " + str(cmdl))
        envp = sm.getEnvironment(p.getID()).get()
        print("Environment: " + str(envp))

if __name__ == '__main__':
    test()



# *****************************************************************************
# * Copyright (c) 2020 Xilinx, Inc.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Xilinx
# *****************************************************************************

from .. import Service, ServiceProvider, add_service_provider
from . import CommandServer
from ..test import NAME, TestListener
from .. import from_xargs
from chipscopy.dm.request import CsRequest, DoneRequest
from chipscopy import dm
from chipscopy.dm import debugcore
from chipscopy.tcf import protocol

MANAGER_TEST_NAME = "test_manager"

listeners = set()


class CsDoneHw(DoneRequest):
    def __init__(self, req):
        self.req = req

    def done_request(self, token, error, args):
        if error:
            self.req.sendError(error)
        else:
            if not args:
                args = ()
            self.req.sendXicomResult(*args)


class TestService(Service):
    def __init__(self):
        pass

    def getName(self):
        return NAME

    def addManagerCmd(self, req):
        args = from_xargs(req.args)
        name = args[0]
        manager = debugcore.DebugCoreManager(name)

        if manager:
            dm.add_manager(manager)
            req.sendXicomResult()
        else:
            req.sendError(f"Could not create manager {name}")

    def removeManagerCmd(self, req):
        args = from_xargs(req.args)

        name = args[0]

        manager = dm.get_manager(name)

        if not manager:
            req.SendError(f"Could not find manager {name}")

        dm.remove_manager(manager)
        req.sendXicomResult()

    def addNodeCmd(self, req):
        args = from_xargs(req.args)

        manager_name = args[0]
        ctx = args[1]
        props = args[2]

        manager = dm.get_manager(manager_name)

        if not manager:
            req.sendError(f"Invalid manager name {manager_name}")
            return

        class AddTestNodeRequest(CsRequest):
            def run(self):
                parent_ctx = props.get("ParentID")
                if not parent_ctx:
                    parent_ctx = ""
                node = manager.add_node(ctx, parent_ctx)

                if node:
                    if props:
                        node.update(props)
                    protocol.invokeLater(req.sendXicomResult)
                else:
                    req.sendError("Could not create node {ctx}, parent {parent_ctx}")
                return True

        cs_req = AddTestNodeRequest(manager)
        cs_req()

    def removeNodeCmd(self, req):
        args = from_xargs(req.args)

        manager_name = args[0]
        ctx = args[1]

        manager = dm.get_manager(manager_name)

        if not manager:
            req.sendError(f"Invalid manager name {manager_name}")
            return

        class RemoveTestNodeRequest(CsRequest):
            def run(self):
                manager.remove_node(ctx)
                protocol.invokeLater(self.done_run)

        cs_req = RemoveTestNodeRequest(manager, done=CsDoneHw(req))
        cs_req()

    def updateNodeCmd(self, req):
        args = from_xargs(req.args)

        manager_name = args[0]
        ctx = args[1]
        props = args[2]

        manager = dm.get_manager(manager_name)

        if not manager:
            req.sendError(f"Invalid manager name {manager_name}")
            return

        class UpdateNodeRequest(CsRequest):
            def run(self):
                self.node.update(props)
                protocol.invokeLater(self.done_run)

        cs_req = UpdateNodeRequest(manager, ctx, done=CsDoneHw(req))
        cs_req()

    def exitCmd(self, req):
        req.sendXicomResult()
        for listener in listeners:
            protocol.invokeLater(listener.exiting)


class TestServiceProvider(ServiceProvider):
    def get_local_service(self, _channel):
        _channel.addCommandServer(
            TestService.service,
            CommandServer(TestService.service, _channel)
        )
        return TestService.service,


# creating singleton
TestService.service = TestService()
add_service_provider(TestServiceProvider())



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from chipscopy.tcf.services import ServiceProvider, add_service_provider
from chipscopy.utils.logger import log


class ProxiesServiceProvider(ServiceProvider):
    package_base = str(__package__)

    def get_local_service(self, channel):
        return ()

    def get_service_proxy(self, channel, service_name: str):
        package_base = str(__package__)
        service = None
        try:
            clsName = service_name + "Proxy"
            package = package_base + "." + clsName
            clsModule = __import__(package, fromlist=[clsName], globals=globals())
            cls = clsModule.__dict__.get(clsName)
            service = cls(channel)
        except ImportError:
            pass
        except Exception as x:
            log.all.error(f"Cannot instantiate service proxy for {service_name}: {x}")
        return service


def init():
    add_service_provider(ProxiesServiceProvider())



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import List
import struct

from chipscopy.client.mem import MemoryNode
from chipscopy import dm
from chipscopy.dm.memory import MemoryManager
from chipscopy.client.view_info import SyncNode, ViewInfo


class Memory(SyncNode):
    """Provides memory read and memory write service for a device.

    Examples: ::

            some_values = device.memory_read(address=0xF2010000, num=2, target="DPC")
            rpu = session.memory.filter_by(name="RPU").at(0)
            some_values = rpu.memory_read(0xF2010000, 2)
            dpc = device.memory.get(name="DPC")
            dpc.memory_write(0xF2010000, [0x12345678, 0xDEADC0FE])
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __str__(self):
        return self.props["Name"]

    def __repr__(self):
        return json.dumps(self.props, indent=4, default=lambda o: str(o))

    @property
    def filter_by(self):
        """Specifies the dict used for filtering memory objects in the device and session.
        Filters memory in the QueryList returned by Session.memory and Device.memory.

        Returns:
              dict[str,value]

        """
        # This is used by the filter_by method in QueryList -
        # add a lowercase copy of all keys for filter convenience
        filter_dict_lower = {k.lower(): v for k, v in self.props.items()}
        filter_dict = {**filter_dict_lower, **self.props.copy()}
        return filter_dict

    def post_init(self):
        """Called at the end of node upgrade"""
        self.manager.get_node(self.ctx, MemoryNode)

    @staticmethod
    def check_and_upgrade(memory_view: ViewInfo, node: MemoryNode) -> "Memory":
        """Converts a client MemoryNode to Memory"""
        if isinstance(node, Memory):
            pass
        elif isinstance(node, MemoryNode) or node.props["node_cls"] == MemoryNode:
            node = memory_view.upgrade_node(node, Memory)
        else:
            raise TypeError("Memory:check_and_upgrade - must be type MemoryNode to upgrade")
        return node

    @staticmethod
    def is_compatible(node: dm.Node) -> bool:
        """True if Memory is a compatible node type"""
        if node and node.manager and node.manager.cs_manager:
            return isinstance(node.manager.cs_manager, MemoryManager)
        else:
            return False

    def memory_write(self, address: int, values: List[int], size: str = "w"):
        """Write values from list of <values> to memory address specified by
        <address>.

        size defaults to 'w' if not specified.

        size <access-size> can be one of the values below:
          |  b = Bytes accesses
          |  h = Half-word accesses
          |  w = Word accesses (default)
          |  d = Double-word accesses

        Args:
            address: Address for memory write
            values: List of integer values to write
            size: 'b','h','w','d'
        """
        bytes_per_word_lookup = {"b": 1, "h": 2, "w": 4, "d": 8}
        bytes_per_word = bytes_per_word_lookup[size]
        byte_value_list = []
        for val in values:
            for i in range(bytes_per_word):
                byte_value_list.append((val & (0xFF << (i * 8))) >> (i * 8))

        # Leading 0 Pad to 32-bit boundary
        required_padding_bytes = (4 - len(byte_value_list)) % 4
        padding = [0 for i in range(required_padding_bytes)]
        byte_value_list = byte_value_list + padding
        packed_word_list = [
            struct.pack("<BBBB", *byte_value_list[i : i + 4])
            for i in range(0, len(byte_value_list), 4)
        ]
        int_word_list = [struct.unpack("<L", word)[0] for word in packed_word_list]
        self.mwr(address=address, words=int_word_list)

    def memory_read(self, address: int, num: int = 1, size: str = "w") -> List[int]:
        """Read and return <num> data values from the target's memory address
        specified by <address>.

        num defaults to 1 if not specified.
        size defaults to 'w' if not specified.

        size <access-size> can be one of the values below:
          |  b = Bytes accesses
          |  h = Half-word accesses
          |  w = Word accesses (default)
          |  d = Double-word accesses

        Args:
            address: Address for memory write
            num: number of values to read
            size: 'b','h','w','d'
        """
        bytes_per_word_lookup = {"b": 1, "h": 2, "w": 4, "d": 8}
        bytes_per_word = bytes_per_word_lookup[size]
        bytes_to_read = num * bytes_per_word
        words_to_read = bytes_to_read // 4
        if (bytes_to_read % 4) != 0:
            words_to_read += 1
        words_read = self.mrd(address, words_to_read)
        result_byte_array = None
        for word in words_read:
            if result_byte_array is None:
                result_byte_array = bytearray(word.to_bytes(4, byteorder="little", signed=False))
            else:
                result_byte_array += bytearray(word.to_bytes(4, byteorder="little", signed=False))
        result_byte_list = list(result_byte_array)[:bytes_to_read]
        result = [0] * num
        word_index = 0
        for byte_index in range(len(result_byte_list)):
            the_byte = result_byte_list[byte_index]
            shift_left_amount = (byte_index % bytes_per_word) * 8
            result[word_index] |= the_byte << shift_left_amount
            if (byte_index + 1) % bytes_per_word == 0:
                word_index += 1
        return result



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TypeVar, Generic, TYPE_CHECKING, Dict, Any, Optional
from dataclasses import dataclass, field

from chipscopy.api import CoreInfo, get_core_info
from chipscopy.api._detail.property import PropertyCommands

if TYPE_CHECKING:
    from chipscopy import CoreType


T = TypeVar("T")


@dataclass
class DebugCore(Generic[T]):
    """DebugCore maps a CoreType to the associated tcf_node and core_implementation.

    At the low level, debug cores have a tcf node that is used to get and set properties.
    This high level API hides the tcf nodes by wrapping them with a core implementation.
    During discover_and_setup_cores() core scanning, the device.debug_core_dict
     is set up with CoreType -> _DebugCore mapping for every detected core.
    """

    core_type: "CoreType"
    core_tcf_node: T
    filter_by: Dict[str, Any] = field(default_factory=dict)
    property: PropertyCommands = None
    core_info: Optional[CoreInfo] = None

    def __post_init__(self):
        self.property = PropertyCommands(self)
        self.core_info = get_core_info(self.core_tcf_node)

    # NOTE - If you want to use the @property to define a property for this class, make sure to
    #  use @builtins.property and not just @property.
    #  This is needed since this class defines an instance var with the same name.



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Dict, Union

from chipscopy.api.ibert.aliases import (
    ALIAS_DICT,
    CHILDREN,
    HANDLE_NAME,
    MODIFIABLE_ALIASES,
    PLL_SOURCE,
    PROPERTY_ENDPOINT,
    RX_KEY,
    TX_KEY,
    TYPE,
)
from chipscopy.api.ibert.rx import RX
from chipscopy.api.ibert.serial_object_base import SerialObjectBase
from chipscopy.api.ibert.tx import TX
from more_itertools import one
from typing_extensions import final

if TYPE_CHECKING:  # pragma: no cover
    from chipscopy.api.ibert.gt_group import GTGroup
    from chipscopy.api.ibert.pll import PLL


@final
class GT(SerialObjectBase["GTGroup", Union[TX, RX]]):
    def __init__(self, name, parent, tcf_node, configuration):
        SerialObjectBase.__init__(
            self,
            name=name,
            type=configuration[TYPE],
            parent=parent,
            handle=configuration[HANDLE_NAME],
            core_tcf_node=tcf_node,
            property_for_alias=configuration.get(ALIAS_DICT, dict()),
            is_property_endpoint=configuration.get(PROPERTY_ENDPOINT, False),
            modifiable_aliases=configuration.get(MODIFIABLE_ALIASES, set()),
        )

        # This is used by the filter_by method in QueryList
        self.filter_by = {"name": self.name, "type": self.type, "handle": self.handle}

        if CHILDREN not in configuration:
            return

        # Create this before creating TX and RX. They will store a reference to this
        self.pll: PLL = None
        """PLL driving the GT"""

        # Build the child objects
        for child_name, child_configuration in configuration[CHILDREN].items():
            if child_configuration[TYPE] == RX_KEY:
                child_class = RX
            elif child_configuration[TYPE] == TX_KEY:
                child_class = TX
            else:
                continue

            new_child = child_class(
                name=child_name,
                parent=self,
                tcf_node=self.core_tcf_node,
                configuration=child_configuration,
            )
            self.children.append(new_child)

        # We always expect a GT to have one valid RX and one valid TX, no more, no less.
        self.rx: RX = one(
            self.children.filter_by(type=RX_KEY),
            too_short=RuntimeError(f"No RX associated with GT {self.handle}!"),
            too_long=RuntimeError(f"More than one RX found for GT {self.handle}!"),
        )
        """RX of the GT"""

        self.tx: TX = one(
            self.children.filter_by(type=TX_KEY),
            too_short=RuntimeError(f"No TX associated with GT {self.handle}!"),
            too_long=RuntimeError(f"More than one TX found for GT {self.handle}!"),
        )
        """TX of the GT"""

    def __repr__(self) -> str:
        return self.name

    def update_pll(self):
        if PLL_SOURCE not in self.property_for_alias:
            return

        _, pll_name_for_this_gt = self.property.get(self.property_for_alias[PLL_SOURCE]).popitem()
        for pll in self.parent.plls:
            if pll.name == pll_name_for_this_gt:
                self.pll = pll
                break

    def reset(self):
        """
        Reset the TX and RX in the GT
        """
        if self.rx:
            self.rx.reset()
        if self.tx:
            self.tx.reset()



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import Optional
from dataclasses import dataclass, InitVar

from chipscopy.utils import printer
from chipscopy.client import ServerInfo
from chipscopy.vivado_version import __vivado_version__


class ServerVersionInfo:
    """ServerVersionInfo tracks the connected build and version information
    for a connected hw_server or cs_server.
    """

    url: str
    build: str
    version: str
    timestamp: str
    server_type: str
    package: Optional[str] = None
    artifact: Optional[str] = None

    def __init__(self, server: ServerInfo, server_type: str):
        self.server_type = server_type
        if server_type == "hw_server":
            xicom_service = server.get_sync_service("Xicom")
            self.url = server.url
            version_dict = xicom_service.get_hw_server_version_info().get()
            self.build = version_dict["build_number"]
            self.version = version_dict["version"]
            self.timestamp = version_dict["timestamp"]
        elif server_type == "cs_server":
            cs_service = server.get_sync_service("ChipScope")
            self.url = server.url
            version_info = cs_service.get_chipscope_server_version_info().args
            if isinstance(version_info, str):
                version_dict = json.loads(version_info)
            elif isinstance(version_info, dict):
                version_dict = version_info
            else:
                raise TypeError(
                    f"Unsupported return type from getChipscopeServerVersionInfo {type(version_info)}"
                )
            self.build = version_dict["build_number"]
            self.version = version_dict["version"]
            self.timestamp = version_dict["timestamp"]
            self.package = version_dict["package_version"]
            self.artifact = version_dict["artifact_type"]
        else:
            raise (ValueError("Invalid server_type - must be hw_server or cs_server"))

    def __str__(self):
        retval = (
            f"\tBuild        : {self.build}\n"
            f"\tVersion      : {self.version}\n"
            f"\tTimestamp    : {self.timestamp}\n"
        )

        if self.server_type == "cs_server":
            retval += f"\tPackage      : {self.package}\n"
            retval += f"\tArtifact Type: {self.artifact}"

        return retval

    def __repr__(self):
        retval = (
            f"{type(self).__name__}("
            f"server_type='{self.server_type}', "
            f"version='{self.version}', timestamp='{self.timestamp}', build='{self.build}', "
            f"package='{self.package}', artifact='{self.artifact}'"
            f")"
        )
        return retval


@dataclass
class VersionDetails:
    version: InitVar[str]
    year: int = 0
    major: int = 0
    minor: int = 0

    def __repr__(self):
        return f"{self.year}.{self.major}.{self.minor}"

    def __post_init__(self, version: str):
        split_data = version.split(".")
        self.year = int(split_data[0])
        self.major = int(split_data[1])
        self.minor = 0 if len(split_data) == 2 else split_data[2]


def version_consistency_check(
    hw_server: ServerInfo, cs_server: Optional[ServerInfo], bypass_version_check: bool
):
    """
    This function will look at the version triplet for ChipScoPy, hw_server and
    cs_server(if applicable), compare their year and major fields i.e. for
    version triplet "2020.2.3", year = 2020 and major = 2 and emit an error if they
    aren't equal.

    The user can chose to bypass this error message and continue using the APIs by
    providing 'bypass_version_check=True' while creating the session.
    """
    mismatch_detected = False
    chipscopy_version = VersionDetails(__vivado_version__)
    hw_server_version = VersionDetails(ServerVersionInfo(hw_server, "hw_server").version)

    # First compare chipscopy and hw_server
    if (
        chipscopy_version.year != hw_server_version.year
        or chipscopy_version.major != hw_server_version.major
    ):
        mismatch_detected = True

    cs_server_version = None
    if cs_server:
        cs_server_version = VersionDetails(ServerVersionInfo(cs_server, "cs_server").version)

        # Next compare cs_server with chipscopy and hw_server
        if (
            chipscopy_version.year != cs_server_version.year
            or chipscopy_version.major != cs_server_version.major
            or hw_server_version.year != cs_server_version.year
            or hw_server_version.major != cs_server_version.major
        ):
            mismatch_detected = True

    if mismatch_detected:
        msg = (
            "A client/server version mismatch was detected while connecting!"
            f"\n   ChipScoPy client version - {chipscopy_version}"
            f"\n   hw_server version        - {hw_server_version}"
        )
        if cs_server_version is not None:
            msg += f"\n   cs_server version        - {cs_server_version}"

        if not bypass_version_check:
            msg += (
                "\nResolution - Verify that your ChipScoPy client matches "
                "the hw_server and cs_server(if applicable) and connect again."
                "\nIf you would like to continue using your ChipScoPy client with the "
                "mismatched server(s), please pass "
                'kwarg "bypass_version_check=True" while creating the session.\n'
            )
            raise RuntimeError(msg)

        printer(msg, level="warning")



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from dataclasses import fields
from typing import Tuple


def dataclass_fields(cls: type) -> Tuple[str]:
    """Make tuple with data class member names."""
    return tuple([f.name for f in fields(cls)])


def filter_props(props: {}, prop_filter: [str]) -> {}:
    return {n: v for n, v in props.items() if n in prop_filter}



# *****************************************************************************
# * Copyright (c) 2011, 2013, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

from .. import protocol, errors, services
from ..channel import Token, toJSONSequence, fromJSONSequence, dumpJSONObject


class Command(object):
    """This is utility class that helps to implement sending a command and
    receiving command result over TCF communication channel.

    The class uses JSON to encode command arguments and to decode result data.

    The class also provides support for TCF standard error report encoding.

    Clients are expected to subclass <code>Command</code> and override
    <code>done</code> method.

    Note: most clients don't need to handle protocol commands directly and can
    use service APIs instead. Service API does all command encoding/decoding
    for a client.

    Typical usage example:

    def getContext(self, id, done):
        class GetContextCommand(Command):
            def done(self, error, args):
                ctx = None
                if not error:
                    assert len(args) == 2
                    error = self.toError(args[0])
                    if args[1]: ctx = Context(args[1])
                done.doneGetContext(self.token, error, ctx)
         command = GetContextCommand(self.channel, self, "getContext", [id])
         return command.token
    """
    __done = False

    def __init__(self, channel, service, command, args):
        if isinstance(service, services.Service):
            service = service.getName()
        self.service = service
        self.command = command
        self.args = args
        t = None
        try:
            # TODO zero_copy
            # zero_copy = channel.isZeroCopySupported()
            t = channel.sendCommand(service, command, toJSONSequence(args),
                                    self)
        except Exception as y:
            t = Token()
            protocol.invokeLater(self._error, y)
        self.token = t

    def _error(self, error):
        assert not self.__done
        self.__done = True
        self.done(error, None)

    def progress(self, token, data):
        assert self.token is token
        error = None
        args = None
        try:
            args = fromJSONSequence(data)
        except Exception as e:
            error = e
        assert not self.__done
        self.progress_update(error, args)

    def result(self, token, data):
        assert self.token is token
        error = None
        args = None
        try:
            args = fromJSONSequence(data)
        except Exception as e:
            error = e
        assert not self.__done
        self.__done = True
        self.done(error, args)

    def terminated(self, token, error):
        assert self.token is token
        assert not self.__done
        self.__done = True
        self.done(error, None)

    def done(self, error, args):
        raise NotImplementedError("Abstract method")

    def getCommandString(self):
        buf = str(self.service) + ' ' + str(self.command)
        if self.args is not None:
            i = 0
            for arg in self.args:
                if i == 0:
                    buf += " "
                else:
                    buf += ", "
                i += 1
                try:
                    buf += dumpJSONObject(arg)
                except Exception as x:
                    # Exception.message does not exist in python3, better use
                    # str(Exception)
                    buf += '***' + str(x) + '***'
        return buf

    def toError(self, data, include_command_text=True):
        if not isinstance(data, dict):
            return None
        errMap = data
        bf = 'TCF error report:\n'
        if include_command_text:
            cmd = self.getCommandString()
            if len(cmd) > 120:
                cmd = cmd[:120] + "..."
            bf += 'Command: ' + str(cmd)
        bf += errors.appendErrorProps(errMap)
        return errors.ErrorReport(bf, errMap)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import NewType, Callable, Dict
from chipscopy.proxies import DebugCorePollingProxy as proxy
from chipscopy.dm import request
from chipscopy.tcf import protocol

PollEventHandler = NewType("PollEventHandler", Callable[[str, Dict], None])


class DebugCorePollScheduler(object):
    """
    Handles scheduling of DebugCore polls and maintains list of current available polls.

    Iterate through list of current polls via get_polls(<parent_node_ctx>)

    It's best to retain the poll_ctx and access the poll description through the scheduler.

    .. method:: poll = scheduler[poll_ctx]

        Gets poll for a given context id
    """

    def __init__(self, channel):
        assert protocol.isDispatchThread()
        self.serv = channel.getRemoteService(proxy.NAME)
        self.polls = {}
        self.nodes = {}
        scheduler = self

        class DebugCorePollingListener(proxy.DebugCorePollingListener):
            def poll_added(self, props: proxy.DebugCorePoll):
                scheduler._add_poll(props)

            def poll_removed(self, poll_ctx: str):
                scheduler._remove_poll(poll_ctx)

            def poll_event(self, poll_ctx: str, props: Dict):
                poll = scheduler.polls.get(poll_ctx)
                if poll and "event_listeners" in poll:
                    for event_listener in poll["event_listeners"]:
                        event_listener(poll, props)

        self.serv.add_listener(DebugCorePollingListener())

    def _add_poll(self, poll):
        poll_id = poll.get("ID")
        parent_id = poll.get("parentID")
        if poll_id:
            self.polls[poll["ID"]] = poll
            if parent_id:
                if parent_id not in self.nodes:
                    self.nodes[parent_id] = set()
                self.nodes[parent_id].add(poll_id)

    def _remove_poll(self, poll_ctx):
        if poll_ctx in self.polls:
            parent_id = self.polls[poll_ctx].get("parentID")
            if parent_id:
                self.nodes[parent_id].remove(poll_ctx)
            del self.polls[poll_ctx]

    def _add_event_listener(self, poll_id, event_handler: PollEventHandler):
        poll = self.polls[poll_id]
        if "event_listeners" not in poll:
            poll["event_listeners"] = set()
        poll["event_listeners"].add(event_handler)

    def _remove_event_listener(self, poll_id, event_handler: PollEventHandler):
        poll = self.polls[poll_id]
        if "event_listeners" in poll:
            try:
                poll["even_listeners"].remove(event_handler)
            except KeyError:
                pass

    def _retain_poll(self, poll_id):
        poll = self.polls[poll_id]
        if "ref_count" not in poll:
            poll["ref_count"] = 1
        else:
            poll["ref_count"] += 1

    def _release_poll(self, poll_id):
        assert self._get_ref_count(poll_id) > 0
        poll = self.polls[poll_id]
        poll["ref_count"] -= 1
        return poll["ref_count"]

    def _get_ref_count(self, poll_id):
        poll = self.polls[poll_id]
        return poll.get("ref_count", 0)

    def update_polls(self, node_id: str = "", done: request.DoneCallback = None) -> request.Token:
        pending = set()
        ret_token = None

        def done_poll_update(token, error, poll):
            pending.remove(token)
            self._add_poll(poll)
            if not pending and done:
                if node_id:
                    results = self.nodes[node_id]
                else:
                    results = self.polls.keys()
                done(ret_token, error, results)

        def done_get_polls(token, error, poll_ids):
            pending.remove(token)
            if not error:
                for poll_id in poll_ids:
                    pending.add(self.serv.get_context(poll_id, done_poll_update))
            if not pending and done:
                done(ret_token, None, [])

        ret_token = self.serv.get_polls(node_id, done_get_polls)
        pending.add(ret_token)
        return ret_token

    def __getitem__(self, item):
        return self.polls[item]

    def get_polls(self, node_id: str = None) -> proxy.DebugCorePoll:
        """
        Poll iterator given a parent debug core node id

        :param node_id: DebugCore context id
        :return: iterator of polls
        """
        if node_id:
            if node_id not in self.nodes:
                raise Exception("Invalid Node Context")
            for poll_id in self.nodes[node_id]:
                yield self.polls[poll_id]
        else:
            for poll in self.polls.values():
                yield poll

    def get_poll_name(self, poll_id: str) -> str:
        if poll_id in self.polls:
            poll = self.polls[poll_id]
            poll_name = poll["name"]
            return poll_name
        else:
            raise Exception(f"Poll ID cannot be found: {poll_id}")

    def schedule_poll(
        self,
        node_ctx: str,
        seq: proxy.DebugCoreSequence,
        name: str = "",
        period: int = 1000,
        event_handler: PollEventHandler = None,
        done: request.DoneCallback = None,
    ) -> request.Token:
        """
        Schedules a poll with the DebugCorePolling Service

        :param node_ctx: Node context to poll
        :param seq: DebugCore sequence to run at each poll
        :param name: Name of the poll
        :param period: Minium period between polls (ms)
        :param event_handler: Event handler when poll conditions are met
        :param done: Done handler called when poll is registered
        :return: Token for poll registration
        """

        def done_add_poll(token, error, poll_id):
            if not error:
                if event_handler:
                    self._add_event_listener(poll_id, event_handler)
                self._retain_poll(poll_id)
            if done:
                done(token, error, poll_id)

        poll = {"name": name, "period": period, "seq": seq}
        return self.serv.add_poll(node_ctx, poll, done_add_poll)

    def release_poll(
        self,
        poll_ctx: str,
        event_handler: PollEventHandler = None,
        done: request.DoneCallback = None,
    ) -> request.Token:
        """
        Removes a scheduled poll.

        :param poll_ctx: Poll context id
        :param event_handler: Event handler to release.
        :param done: Done handler when complete
        :return: Token for removal
        """
        if event_handler:
            self._remove_event_listener(poll_ctx, event_handler)

        ref_count = self._release_poll(poll_ctx)

        if not ref_count:
            return self.serv.remove_poll(poll_ctx, done)

        t = request.Token()

        if done:
            protocol.invokeLater(done, t, None, None)

        return t

    def retain_poll(
        self,
        poll_ctx: str,
        event_handler: PollEventHandler = None,
        done: request.DoneCallback = None,
    ) -> request.Token:
        """
        Retains a poll so it's not lost if released elsewhere.  Also, this subscribes a new PollEventHandler if
        given.

        :param poll_ctx: Poll context ID
        :param event_handler: Event handler that responds to a poll event
        :param done: Done retain callback
        :return: Token of request
        """

        assert poll_ctx in self.polls

        def done_retain(token, error, result):
            if not error:
                if event_handler:
                    self._add_event_listener(poll_ctx, event_handler)
                self._retain_poll(poll_ctx)
            if done:
                done(token, error, result)

        if self._get_ref_count(poll_ctx):
            # This scheduler already has subscribed to the hw_server so no need to send a TCF message
            t = request.Token()
            if done:
                protocol.invokeLater(done_retain, t, None, None)
            return t

        return self.serv.retain_poll(poll_ctx, done_retain)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from chipscopy.api.ila.ila_capture import (
    ILACaptureCondition,
    ILATriggerCondition,
    ILATrigInMode,
    ILATrigOutMode,
    ILAControl,
    ILA_TRIGGER_POSITION_HALF,
    ILA_WINDOW_SIZE_MAX,
    ILAState,
    ILAStatus,
    ILA_CONTROL_MEMBERS,
    ILA_STATUS_MEMBERS,
)
from chipscopy.api.ila.ila_probe import (
    ILAMatchUnitType,
    ILAPort,
    ILA_PORT_MEMBERS,
    ILAProbe,
    ILA_MATCH_OPERATORS,
    ILA_MATCH_BIT_VALUES,
    ILAProbeValues,
    ILABitRange,
    ILAProbeRadix,
)
from chipscopy.api.ila.ila_waveform import ILAWaveform, ILAWaveformProbe
from chipscopy.api.ila.ila import (
    ILA,
    ILAStaticInfo,
    export_waveform,
    get_waveform_data,
    get_waveform_probe_data,
)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import Enum, EnumMeta
import inspect
import json
import struct
from typing import List, Any, Optional

from chipscopy.utils.printer import printer

WORD_SIZE = 4


def words_from_bytes(buf):
    words = []
    s = struct.Struct("<I")
    for i in range(0, len(buf), WORD_SIZE):
        words.append(s.unpack(buf[i : i + WORD_SIZE])[0])
    return words


def bytes_from_words(words, word_count=0):
    if not word_count:
        word_count = len(words)
    buf = bytearray(word_count * WORD_SIZE)
    count = min(len(words), word_count)
    s = struct.Struct("<I")
    start = 0
    for i in range(count):
        buf[start : start + WORD_SIZE] = s.pack(words[i])
        start += WORD_SIZE

    word_count -= count
    while word_count > 0:
        buf[start : start + WORD_SIZE] = s.pack(words[-1])
        start += WORD_SIZE
        word_count -= 1

    return buf


def word_align(addr: int) -> int:
    return int(addr / WORD_SIZE) * WORD_SIZE


def listify(item) -> List[Any]:
    if isinstance(item, list):
        return item
    elif isinstance(item, str) or isinstance(item, int) or isinstance(item, float):
        return [item]
    else:
        return list(item)


def deprecated_api(*, release: str, replacement: Optional[str] = None):
    def inner(function_to_decorate):
        def actual_decorator(*args, **kwargs):
            deprecated_func_name = list(
                filter(
                    lambda entry: entry[0] == "__name__", inspect.getmembers(function_to_decorate)
                )
            ).pop()[1]
            called_from = inspect.stack()[1]

            msg = f"{deprecated_func_name}() has been deprecated and will be removed in {release}\n"
            if replacement:
                msg += f"Please consider using the alternative {replacement}\n"
            msg += f"\nCalled from line {called_from.lineno} in file {called_from.filename}"

            printer(msg, level="warning")

            return function_to_decorate(*args, **kwargs)

        return actual_decorator

    return inner


class Enum2StrEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Enum):
            return str(obj)
        if isinstance(obj, EnumMeta):
            items = {item.name: item.value for item in list(obj)}
            return items
        return json.JSONEncoder.default(self, obj)
